## @file
# Module that encodes and decodes a EFI_CAPSULE_HEADER with a payload
#
# Copyright (c) 2018, Intel Corporation. All rights reserved.<BR>
# SPDX-License-Identifier: BSD-2-Clause-Patent
#
"""Module for encoding and decoding a EFI_CAPSULE_HEADER with a payload."""

import struct
import uuid
from io import BytesIO

from edk2toollib.uefi.fmp_capsule_header import FmpCapsuleHeaderClass


class UefiCapsuleHeaderClass(object):
    r"""An object representing a UEFI_CAPSULE_HEADER.

    Attributes:
        CapsuleGuid (uuid.UUID):    6DCBD5ED-E82D-4C44-BDA1-7194199AD92A
        HeaderSize (int):           The size of the capsule header. This may be larger than the size of the
                                    EFI_CAPSULE_HEADER since CapsuleGuid may imply extended header entries
        OemFlags (int):             Bit-mapped list describing the capsule attributes. The Flag values of 0x0000 -
                                    0xFFFF are defined by CapsuleGuid. Flag values of 0x10000 - 0xFFFFFFFF are defined
                                    by this specification
        PersistAcrossReset (bool):  Flag pulled from OemFlags
        PopulateSystemTable (bool): Flag pulled from OemFlags
        InitiateReset (bool):       Flag pulled from OemFlags
        Payload (bytes):              string representing packed data as bytes (i.e. b'\x01\x00\x03')
        FmpCapsuleHeader (FmpCapsuleHeaderClass): Fmp Capsule Header

    ```
    typedef struct {
        EFI_GUID          CapsuleGuid;
        UINT32            HeaderSize;
        UINT32            Flags;
        UINT32            CapsuleImageSize;
    } EFI_CAPSULE_HEADER;

    #define CAPSULE_FLAGS_PERSIST_ACROSS_RESET          0x00010000
    #define CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE         0x00020000
    #define CAPSULE_FLAGS_INITIATE_RESET                0x00040000
    ```
    """

    _StructFormat = "<16sIIII"
    _StructSize = struct.calcsize(_StructFormat)

    EFI_FIRMWARE_MANAGEMENT_CAPSULE_ID_GUID = uuid.UUID("6DCBD5ED-E82D-4C44-BDA1-7194199AD92A")

    _CAPSULE_FLAGS_PERSIST_ACROSS_RESET = 0x00010000
    _CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE = 0x00020000
    _CAPSULE_FLAGS_INITIATE_RESET = 0x00040000

    def __init__(self) -> "UefiCapsuleHeaderClass":
        """Inits an empty object."""
        self.CapsuleGuid = self.EFI_FIRMWARE_MANAGEMENT_CAPSULE_ID_GUID
        self.HeaderSize = self._StructSize
        self.OemFlags = 0x0000
        self.PersistAcrossReset = False
        self.PopulateSystemTable = False
        self.InitiateReset = False
        self.CapsuleImageSize = self.HeaderSize
        self.Payload = b""
        self.FmpCapsuleHeader = None

    def Encode(self) -> bytes:
        r"""Serializes the Header + payload.

        Returns:
            (bytes): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        Flags = self.OemFlags
        if self.PersistAcrossReset:
            Flags = Flags | self._CAPSULE_FLAGS_PERSIST_ACROSS_RESET
        if self.PopulateSystemTable:
            Flags = Flags | self._CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE
        if self.InitiateReset:
            Flags = Flags | self._CAPSULE_FLAGS_INITIATE_RESET

        # If we have an FmpCapsuleHeader, let's collapse that now.
        if self.FmpCapsuleHeader is not None:
            self.Payload = self.FmpCapsuleHeader.Encode()

        self.CapsuleImageSize = self.HeaderSize + len(self.Payload)

        UefiCapsuleHeader = struct.pack(
            self._StructFormat, self.CapsuleGuid.bytes_le, self.HeaderSize, Flags, self.CapsuleImageSize, 0
        )

        return UefiCapsuleHeader + self.Payload

    def Decode(self, Buffer: BytesIO) -> bytes:
        """Loads data into the Object by parsing a buffer.

        Args:
            Buffer (obj): Buffer containing the data

        Returns:
            (str): string of binary representing the payload

        Raises:
            (ValueError): Invalid Buffer
            (ValueError): Invalid Signature
            (ValueError): Invalid Header size
        """
        if len(Buffer) < self._StructSize:
            raise ValueError
        (CapsuleGuid, HeaderSize, Flags, CapsuleImageSize, Reserved) = struct.unpack(
            self._StructFormat, Buffer[0 : self._StructSize]
        )
        if HeaderSize < self._StructSize:
            raise ValueError
        if CapsuleImageSize != len(Buffer):
            raise ValueError
        self.CapsuleGuid = uuid.UUID(bytes_le=CapsuleGuid)
        self.HeaderSize = HeaderSize
        self.OemFlags = Flags & 0xFFFF
        self.PersistAcrossReset = (Flags & self._CAPSULE_FLAGS_PERSIST_ACROSS_RESET) != 0
        self.PopulateSystemTable = (Flags & self._CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE) != 0
        self.InitiateReset = (Flags & self._CAPSULE_FLAGS_INITIATE_RESET) != 0
        self.CapsuleImageSize = CapsuleImageSize
        self.Payload = Buffer[self.HeaderSize :]
        if len(self.Payload) > 0 and self.CapsuleGuid == self.EFI_FIRMWARE_MANAGEMENT_CAPSULE_ID_GUID:
            self.FmpCapsuleHeader = FmpCapsuleHeaderClass()
            self.FmpCapsuleHeader.Decode(self.Payload)

        return self.Payload

    def DumpInfo(self) -> None:
        """Prints payload header information."""
        Flags = self.OemFlags
        if self.PersistAcrossReset:
            Flags = Flags | self._CAPSULE_FLAGS_PERSIST_ACROSS_RESET
        if self.PopulateSystemTable:
            Flags = Flags | self._CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE
        if self.InitiateReset:
            Flags = Flags | self._CAPSULE_FLAGS_INITIATE_RESET
        print("EFI_CAPSULE_HEADER.CapsuleGuid      = {Guid}".format(Guid=str(self.CapsuleGuid).upper()))
        print("EFI_CAPSULE_HEADER.HeaderSize       = {Size:08X}".format(Size=self.HeaderSize))
        print("EFI_CAPSULE_HEADER.Flags            = {Flags:08X}".format(Flags=Flags))
        print("  OEM Flags                         = {Flags:04X}".format(Flags=self.OemFlags))
        if self.PersistAcrossReset:
            print("  CAPSULE_FLAGS_PERSIST_ACROSS_RESET")
        if self.PopulateSystemTable:
            print("  CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE")
        if self.InitiateReset:
            print("  CAPSULE_FLAGS_INITIATE_RESET")
        print("EFI_CAPSULE_HEADER.CapsuleImageSize = {Size:08X}".format(Size=self.CapsuleImageSize))
        print("sizeof (Payload)                    = {Size:08X}".format(Size=len(self.Payload)))
        if self.FmpCapsuleHeader is not None:
            self.FmpCapsuleHeader.DumpInfo()



##
# junit_report_format
# This module contains support for Outputting Junit test results xml.
#
# Used to support CI/CD and exporting test results for other tools.
# This does test report generation without being a test runner.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module for outputting Junit test results to xml.

Used to support CI/CD and exporting test results for other tools.
This does test report generation without being a test runner.
"""

import time
from typing import IO
from xml.sax.saxutils import escape


class JunitReportError(object):
    """Object representing a Test Error."""

    def __init__(self, type: str, msg: str) -> "JunitReportError":
        """Init the type of error."""
        self.Message = escape(msg.strip(), {'"': "&quot;"})
        self.Type = escape(type.strip(), {'"': "&quot;"})


class JunitReportFailure(object):
    """Object representing a Test Failure."""

    def __init__(self, type: str, msg: str) -> "JunitReportFailure":
        """Init the type of Failure."""
        self.Message = escape(msg.strip(), {'"': "&quot;"})
        self.Type = escape(type.strip(), {'"': "&quot;"})


class JunitReportTestCase(object):
    """Object representing a single test case."""

    NEW = 1
    SKIPPED = 2
    FAILED = 3
    ERROR = 4
    SUCCESS = 5

    def __init__(self, Name: str, ClassName: str) -> "JunitReportTestCase":
        """Init a Test case with it's name and class name."""
        self.Name = escape(Name.strip(), {'"': "&quot;"})
        self.ClassName = escape(ClassName.strip(), {'"': "&quot;"})
        self.Time = 0
        self.Status = JunitReportTestCase.NEW

        self.FailureMsg = None
        self.ErrorMsg = None
        self._TestSuite = None
        self.StdErr = ""
        self.StdOut = ""
        self._StartTime = time.time()

    def SetFailed(self, Msg: str, Type: str) -> None:
        """Sets internal state if the test failed."""
        if self.Status != JunitReportTestCase.NEW:
            raise Exception("Can't Set to failed.  State must be in NEW")
        self.Time = time.time() - self._StartTime
        self.Status = JunitReportTestCase.FAILED
        self.FailureMsg = JunitReportFailure(Type, Msg)

    def SetError(self, Msg: str, Type: str) -> None:
        """Set internal state if the test had an error."""
        if self.Status != JunitReportTestCase.NEW:
            raise Exception("Can't Set to error.  State must be in NEW")
        self.Time = time.time() - self._StartTime
        self.Status = JunitReportTestCase.ERROR
        self.ErrorMsg = JunitReportError(Type, Msg)

    def SetSuccess(self) -> None:
        """Set internal state if the test passed."""
        if self.Status != JunitReportTestCase.NEW:
            raise Exception("Can't Set to success.  State must be in NEW")
        self.Status = JunitReportTestCase.SUCCESS
        self.Time = time.time() - self._StartTime

    def SetSkipped(self) -> None:
        """Set internal state if the test was skipped."""
        if self.Status != JunitReportTestCase.NEW:
            raise Exception("Can't Set to skipped.  State must be in NEW")
        self.Status = JunitReportTestCase.SKIPPED
        self.Time = time.time() - self._StartTime

    def LogStdOut(self, msg: str) -> None:
        """Log to the standard out."""
        self.StdOut += escape(msg.strip()) + "\n "

    def LogStdError(self, msg: str) -> None:
        """Log to the standard err."""
        self.StdErr += escape(msg.strip()) + "\n "

    def Output(self, outstream: IO) -> None:
        """Write the test result to the outstream."""
        outstream.write('<testcase classname="{0}" name="{1}" time="{2}">'.format(self.ClassName, self.Name, self.Time))
        if self.Status == JunitReportTestCase.SKIPPED:
            outstream.write('<skipped type="skipped">')
            outstream.write(self.StdOut)
            outstream.write("</skipped>")
        elif self.Status == JunitReportTestCase.FAILED:
            outstream.write(
                '<failure message="{0}" type="{1}" />'.format(self.FailureMsg.Message, self.FailureMsg.Type)
            )
        elif self.Status == JunitReportTestCase.ERROR:
            outstream.write('<error message="{0}" type="{1}" />'.format(self.ErrorMsg.Message, self.ErrorMsg.Type))
        elif self.Status != JunitReportTestCase.SUCCESS:
            raise Exception(
                "Can't output a testcase {0}.{1} in invalid state {2}".format(self.ClassName, self.Name, self.Status)
            )

        outstream.write("<system-out>" + self.StdOut + "</system-out>")
        outstream.write("<system-err>" + self.StdErr + "</system-err>")
        outstream.write("</testcase>")


class JunitReportTestSuite(object):
    """Object representing the overall test suite.

    Create new suites by using the JunitTestReport Object
    """

    def __init__(self, Name: str, Package: str, Id: id) -> "JunitReportTestSuite":
        """Initialize a new test suite."""
        self.Name = escape(Name.strip(), {'"': "&quot;"})
        self.Package = escape(Package.strip(), {'"': "&quot;"})
        self.TestId = Id
        self.TestCases = []

    def create_new_testcase(self, name: str, classname: str) -> JunitReportTestCase:
        """Create a new test case.

        Returns:
            (JunitReportTestCase): newly created test case
        """
        tc = JunitReportTestCase(name, classname)
        self.TestCases.append(tc)
        tc._TestSuite = self
        return tc

    def Output(self, outstream: IO) -> None:
        """Output the test results to the stream."""
        Errors = 0
        Failures = 0
        Skipped = 0
        Tests = len(self.TestCases)

        for a in self.TestCases:
            if a.Status == JunitReportTestCase.FAILED:
                Failures += 1
            elif a.Status == JunitReportTestCase.ERROR:
                Errors += 1
            elif a.Status == JunitReportTestCase.SKIPPED:
                Skipped += 1

        outstream.write(
            '<testsuite id="{0}" name="{1}" package="{2}" errors="{3}" tests="{4}" '
            'failures="{5}" skipped="{6}">'.format(
                self.TestId, self.Name, self.Package, Errors, Tests, Failures, Skipped
            )
        )

        for a in self.TestCases:
            a.Output(outstream)

        outstream.write("</testsuite>")


class JunitTestReport(object):
    """Object representing a Test Report.

    Top level object test reporting.
    """

    def __init__(self) -> "JunitTestReport":
        """Init an empty test report."""
        self.TestSuites = []

    def create_new_testsuite(self, name: str, package: str) -> JunitReportTestSuite:
        """Create a new test suite.

        Returns:
            (JunitReportTestSuite): newly created testsuite
        """
        id = len(self.TestSuites)
        ts = JunitReportTestSuite(name, package, id)
        self.TestSuites.append(ts)
        return ts

    def Output(self, filepath: str) -> None:
        """Write report to file."""
        f = open(filepath, "w")
        f.write("")
        f.write('<?xml version="1.0" encoding="UTF-8"?>')
        f.write("<testsuites>")
        for a in self.TestSuites:
            a.Output(f)
        f.write("</testsuites>")
        f.close()



# @file tpm2_defs.py
# This file contains utility classes to help interpret definitions from the
# Tpm20.h header file in TianoCore.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""This module contains utility classes to help interpret definitions from the Tpm20.h header file."""

from typing import Optional

##
# INCLUDES CONTENTS FROM TianoCore Tpm20.h HEADER FILE!!
#
# TPM2.0 Specification data structures
#   (Trusted Platform Module Library Specification, Family "2.0", Level 00, Revision 00.96,
#   @http://www.trustedcomputinggroup.org/resources/tpm_library_specification)
#
#   Check http://trustedcomputinggroup.org for latest specification updates.
#
# Copyright (c) 2013 - 2015, Intel Corporation. All rights reserved. <BR>
##

# spell-checker:disable


# Table 7 - TPM_ALG_ID Constants
TPM_ALG_Size = 2
TPM_ALG_Pack = "H"
TPM_ALG_ERROR = 0x0000
TPM_ALG_FIRST = 0x0001
TPM_ALG_RSA = 0x0001
TPM_ALG_SHA = 0x0004
TPM_ALG_SHA1 = 0x0004
TPM_ALG_HMAC = 0x0005
TPM_ALG_AES = 0x0006
TPM_ALG_MGF1 = 0x0007
TPM_ALG_KEYEDHASH = 0x0008
TPM_ALG_XOR = 0x000A
TPM_ALG_SHA256 = 0x000B
TPM_ALG_SHA384 = 0x000C
TPM_ALG_SHA512 = 0x000D
TPM_ALG_NULL = 0x0010
TPM_ALG_SM3_256 = 0x0012
TPM_ALG_SM4 = 0x0013
TPM_ALG_RSASSA = 0x0014
TPM_ALG_RSAES = 0x0015
TPM_ALG_RSAPSS = 0x0016
TPM_ALG_OAEP = 0x0017
TPM_ALG_ECDSA = 0x0018
TPM_ALG_ECDH = 0x0019
TPM_ALG_ECDAA = 0x001A
TPM_ALG_SM2 = 0x001B
TPM_ALG_ECSCHNORR = 0x001C
TPM_ALG_ECMQV = 0x001D
TPM_ALG_KDF1_SP800_56a = 0x0020
TPM_ALG_KDF2 = 0x0021
TPM_ALG_KDF1_SP800_108 = 0x0022
TPM_ALG_ECC = 0x0023
TPM_ALG_SYMCIPHER = 0x0025
TPM_ALG_CTR = 0x0040
TPM_ALG_OFB = 0x0041
TPM_ALG_CBC = 0x0042
TPM_ALG_CFB = 0x0043
TPM_ALG_ECB = 0x0044
TPM_ALG_LAST = 0x0044

# Table 11 - TPM_CC Constants (Numeric Order)
TPM_CC_Size = 4
TPM_CC_Pack = "L"
TPM_CC_FIRST = 0x0000011F
TPM_CC_PP_FIRST = 0x0000011F
TPM_CC_NV_UndefineSpaceSpecial = 0x0000011F
TPM_CC_EvictControl = 0x00000120
TPM_CC_HierarchyControl = 0x00000121
TPM_CC_NV_UndefineSpace = 0x00000122
TPM_CC_ChangeEPS = 0x00000124
TPM_CC_ChangePPS = 0x00000125
TPM_CC_Clear = 0x00000126
TPM_CC_ClearControl = 0x00000127
TPM_CC_ClockSet = 0x00000128
TPM_CC_HierarchyChangeAuth = 0x00000129
TPM_CC_NV_DefineSpace = 0x0000012A
TPM_CC_PCR_Allocate = 0x0000012B
TPM_CC_PCR_SetAuthPolicy = 0x0000012C
TPM_CC_PP_Commands = 0x0000012D
TPM_CC_SetPrimaryPolicy = 0x0000012E
TPM_CC_FieldUpgradeStart = 0x0000012F
TPM_CC_ClockRateAdjust = 0x00000130
TPM_CC_CreatePrimary = 0x00000131
TPM_CC_NV_GlobalWriteLock = 0x00000132
TPM_CC_PP_LAST = 0x00000132
TPM_CC_GetCommandAuditDigest = 0x00000133
TPM_CC_NV_Increment = 0x00000134
TPM_CC_NV_SetBits = 0x00000135
TPM_CC_NV_Extend = 0x00000136
TPM_CC_NV_Write = 0x00000137
TPM_CC_NV_WriteLock = 0x00000138
TPM_CC_DictionaryAttackLockReset = 0x00000139
TPM_CC_DictionaryAttackParameters = 0x0000013A
TPM_CC_NV_ChangeAuth = 0x0000013B
TPM_CC_PCR_Event = 0x0000013C
TPM_CC_PCR_Reset = 0x0000013D
TPM_CC_SequenceComplete = 0x0000013E
TPM_CC_SetAlgorithmSet = 0x0000013F
TPM_CC_SetCommandCodeAuditStatus = 0x00000140
TPM_CC_FieldUpgradeData = 0x00000141
TPM_CC_IncrementalSelfTest = 0x00000142
TPM_CC_SelfTest = 0x00000143
TPM_CC_Startup = 0x00000144
TPM_CC_Shutdown = 0x00000145
TPM_CC_StirRandom = 0x00000146
TPM_CC_ActivateCredential = 0x00000147
TPM_CC_Certify = 0x00000148
TPM_CC_PolicyNV = 0x00000149
TPM_CC_CertifyCreation = 0x0000014A
TPM_CC_Duplicate = 0x0000014B
TPM_CC_GetTime = 0x0000014C
TPM_CC_GetSessionAuditDigest = 0x0000014D
TPM_CC_NV_Read = 0x0000014E
TPM_CC_NV_ReadLock = 0x0000014F
TPM_CC_ObjectChangeAuth = 0x00000150
TPM_CC_PolicySecret = 0x00000151
TPM_CC_Rewrap = 0x00000152
TPM_CC_Create = 0x00000153
TPM_CC_ECDH_ZGen = 0x00000154
TPM_CC_HMAC = 0x00000155
TPM_CC_Import = 0x00000156
TPM_CC_Load = 0x00000157
TPM_CC_Quote = 0x00000158
TPM_CC_RSA_Decrypt = 0x00000159
TPM_CC_HMAC_Start = 0x0000015B
TPM_CC_SequenceUpdate = 0x0000015C
TPM_CC_Sign = 0x0000015D
TPM_CC_Unseal = 0x0000015E
TPM_CC_PolicySigned = 0x00000160
TPM_CC_ContextLoad = 0x00000161
TPM_CC_ContextSave = 0x00000162
TPM_CC_ECDH_KeyGen = 0x00000163
TPM_CC_EncryptDecrypt = 0x00000164
TPM_CC_FlushContext = 0x00000165
TPM_CC_LoadExternal = 0x00000167
TPM_CC_MakeCredential = 0x00000168
TPM_CC_NV_ReadPublic = 0x00000169
TPM_CC_PolicyAuthorize = 0x0000016A
TPM_CC_PolicyAuthValue = 0x0000016B
TPM_CC_PolicyCommandCode = 0x0000016C
TPM_CC_PolicyCounterTimer = 0x0000016D
TPM_CC_PolicyCpHash = 0x0000016E
TPM_CC_PolicyLocality = 0x0000016F
TPM_CC_PolicyNameHash = 0x00000170
TPM_CC_PolicyOR = 0x00000171
TPM_CC_PolicyTicket = 0x00000172
TPM_CC_ReadPublic = 0x00000173
TPM_CC_RSA_Encrypt = 0x00000174
TPM_CC_StartAuthSession = 0x00000176
TPM_CC_VerifySignature = 0x00000177
TPM_CC_ECC_Parameters = 0x00000178
TPM_CC_FirmwareRead = 0x00000179
TPM_CC_GetCapability = 0x0000017A
TPM_CC_GetRandom = 0x0000017B
TPM_CC_GetTestResult = 0x0000017C
TPM_CC_Hash = 0x0000017D
TPM_CC_PCR_Read = 0x0000017E
TPM_CC_PolicyPCR = 0x0000017F
TPM_CC_PolicyRestart = 0x00000180
TPM_CC_ReadClock = 0x00000181
TPM_CC_PCR_Extend = 0x00000182
TPM_CC_PCR_SetAuthValue = 0x00000183
TPM_CC_NV_Certify = 0x00000184
TPM_CC_EventSequenceComplete = 0x00000185
TPM_CC_HashSequenceStart = 0x00000186
TPM_CC_PolicyPhysicalPresence = 0x00000187
TPM_CC_PolicyDuplicationSelect = 0x00000188
TPM_CC_PolicyGetDigest = 0x00000189
TPM_CC_TestParms = 0x0000018A
TPM_CC_Commit = 0x0000018B
TPM_CC_PolicyPassword = 0x0000018C
TPM_CC_ZGen_2Phase = 0x0000018D
TPM_CC_EC_Ephemeral = 0x0000018E
TPM_CC_LAST = 0x0000018E

# Table 18 - TPM_ST Constants
TPM_ST_Size = 2
TPM_ST_Pack = "H"
TPM_ST_RSP_COMMAND = 0x00C4
TPM_ST_NULL = 0x8000
TPM_ST_NO_SESSIONS = 0x8001
TPM_ST_SESSIONS = 0x8002
TPM_ST_ATTEST_NV = 0x8014
TPM_ST_ATTEST_COMMAND_AUDIT = 0x8015
TPM_ST_ATTEST_SESSION_AUDIT = 0x8016
TPM_ST_ATTEST_CERTIFY = 0x8017
TPM_ST_ATTEST_QUOTE = 0x8018
TPM_ST_ATTEST_TIME = 0x8019
TPM_ST_ATTEST_CREATION = 0x801A
TPM_ST_CREATION = 0x8021
TPM_ST_VERIFIED = 0x8022
TPM_ST_AUTH_SECRET = 0x8023
TPM_ST_HASHCHECK = 0x8024
TPM_ST_AUTH_SIGNED = 0x8025
TPM_ST_FU_MANIFEST = 0x8029

# Table 19 - TPM_SU Constants
TPM_SU_Size = 2
TPM_SU_Pack = "H"
TPM_SU_CLEAR = 0x0000
TPM_SU_STATE = 0x0001

# Table 20 - TPM_SE Constants
TPM_SE_Size = 1
TPM_SE_Pack = "B"
TPM_SE_HMAC = 0x00
TPM_SE_POLICY = 0x01
TPM_SE_TRIAL = 0x03

# Table 27 - TPM_RH Constants
TPM_RH_Size = 4
TPM_RH_Pack = "L"
TPM_RH_FIRST = 0x40000000
TPM_RH_SRK = 0x40000000
TPM_RH_OWNER = 0x40000001
TPM_RH_REVOKE = 0x40000002
TPM_RH_TRANSPORT = 0x40000003
TPM_RH_OPERATOR = 0x40000004
TPM_RH_ADMIN = 0x40000005
TPM_RH_EK = 0x40000006
TPM_RH_NULL = 0x40000007
TPM_RH_UNASSIGNED = 0x40000008
TPM_RS_PW = 0x40000009
TPM_RH_LOCKOUT = 0x4000000A
TPM_RH_ENDORSEMENT = 0x4000000B
TPM_RH_PLATFORM = 0x4000000C
TPM_RH_PLATFORM_NV = 0x4000000D
TPM_RH_AUTH_00 = 0x40000010
TPM_RH_AUTH_FF = 0x4000010F
TPM_RH_LAST = 0x4000010F


class CommandCode(object):
    """Class representing a command code.

    Allows alternativing between command codes and command strings.
    """

    @staticmethod
    def get_code(cc_string: str) -> Optional[int]:
        """Returns associated command code.

        Args:
            cc_string (str): command string
        Returns:
            (int): Associated command code
            (None): cc_string not found
        """
        return {
            "TPM_CC_NV_UndefineSpaceSpecial": 0x0000011F,
            "TPM_CC_EvictControl": 0x00000120,
            "TPM_CC_HierarchyControl": 0x00000121,
            "TPM_CC_NV_UndefineSpace": 0x00000122,
            "TPM_CC_ChangeEPS": 0x00000124,
            "TPM_CC_ChangePPS": 0x00000125,
            "TPM_CC_Clear": 0x00000126,
            "TPM_CC_ClearControl": 0x00000127,
            "TPM_CC_ClockSet": 0x00000128,
            "TPM_CC_HierarchyChangeAuth": 0x00000129,
            "TPM_CC_NV_DefineSpace": 0x0000012A,
            "TPM_CC_PCR_Allocate": 0x0000012B,
            "TPM_CC_PCR_SetAuthPolicy": 0x0000012C,
            "TPM_CC_PP_Commands": 0x0000012D,
            "TPM_CC_SetPrimaryPolicy": 0x0000012E,
            "TPM_CC_FieldUpgradeStart": 0x0000012F,
            "TPM_CC_ClockRateAdjust": 0x00000130,
            "TPM_CC_CreatePrimary": 0x00000131,
            "TPM_CC_NV_GlobalWriteLock": 0x00000132,
            "TPM_CC_GetCommandAuditDigest": 0x00000133,
            "TPM_CC_NV_Increment": 0x00000134,
            "TPM_CC_NV_SetBits": 0x00000135,
            "TPM_CC_NV_Extend": 0x00000136,
            "TPM_CC_NV_Write": 0x00000137,
            "TPM_CC_NV_WriteLock": 0x00000138,
            "TPM_CC_DictionaryAttackLockReset": 0x00000139,
            "TPM_CC_DictionaryAttackParameters": 0x0000013A,
            "TPM_CC_NV_ChangeAuth": 0x0000013B,
            "TPM_CC_PCR_Event": 0x0000013C,
            "TPM_CC_PCR_Reset": 0x0000013D,
            "TPM_CC_SequenceComplete": 0x0000013E,
            "TPM_CC_SetAlgorithmSet": 0x0000013F,
            "TPM_CC_SetCommandCodeAuditStatus": 0x00000140,
            "TPM_CC_FieldUpgradeData": 0x00000141,
            "TPM_CC_IncrementalSelfTest": 0x00000142,
            "TPM_CC_SelfTest": 0x00000143,
            "TPM_CC_Startup": 0x00000144,
            "TPM_CC_Shutdown": 0x00000145,
            "TPM_CC_StirRandom": 0x00000146,
            "TPM_CC_ActivateCredential": 0x00000147,
            "TPM_CC_Certify": 0x00000148,
            "TPM_CC_PolicyNV": 0x00000149,
            "TPM_CC_CertifyCreation": 0x0000014A,
            "TPM_CC_Duplicate": 0x0000014B,
            "TPM_CC_GetTime": 0x0000014C,
            "TPM_CC_GetSessionAuditDigest": 0x0000014D,
            "TPM_CC_NV_Read": 0x0000014E,
            "TPM_CC_NV_ReadLock": 0x0000014F,
            "TPM_CC_ObjectChangeAuth": 0x00000150,
            "TPM_CC_PolicySecret": 0x00000151,
            "TPM_CC_Rewrap": 0x00000152,
            "TPM_CC_Create": 0x00000153,
            "TPM_CC_ECDH_ZGen": 0x00000154,
            "TPM_CC_HMAC": 0x00000155,
            "TPM_CC_Import": 0x00000156,
            "TPM_CC_Load": 0x00000157,
            "TPM_CC_Quote": 0x00000158,
            "TPM_CC_RSA_Decrypt": 0x00000159,
            "TPM_CC_HMAC_Start": 0x0000015B,
            "TPM_CC_SequenceUpdate": 0x0000015C,
            "TPM_CC_Sign": 0x0000015D,
            "TPM_CC_Unseal": 0x0000015E,
            "TPM_CC_PolicySigned": 0x00000160,
            "TPM_CC_ContextLoad": 0x00000161,
            "TPM_CC_ContextSave": 0x00000162,
            "TPM_CC_ECDH_KeyGen": 0x00000163,
            "TPM_CC_EncryptDecrypt": 0x00000164,
            "TPM_CC_FlushContext": 0x00000165,
            "TPM_CC_LoadExternal": 0x00000167,
            "TPM_CC_MakeCredential": 0x00000168,
            "TPM_CC_NV_ReadPublic": 0x00000169,
            "TPM_CC_PolicyAuthorize": 0x0000016A,
            "TPM_CC_PolicyAuthValue": 0x0000016B,
            "TPM_CC_PolicyCommandCode": 0x0000016C,
            "TPM_CC_PolicyCounterTimer": 0x0000016D,
            "TPM_CC_PolicyCpHash": 0x0000016E,
            "TPM_CC_PolicyLocality": 0x0000016F,
            "TPM_CC_PolicyNameHash": 0x00000170,
            "TPM_CC_PolicyOR": 0x00000171,
            "TPM_CC_PolicyTicket": 0x00000172,
            "TPM_CC_ReadPublic": 0x00000173,
            "TPM_CC_RSA_Encrypt": 0x00000174,
            "TPM_CC_StartAuthSession": 0x00000176,
            "TPM_CC_VerifySignature": 0x00000177,
            "TPM_CC_ECC_Parameters": 0x00000178,
            "TPM_CC_FirmwareRead": 0x00000179,
            "TPM_CC_GetCapability": 0x0000017A,
            "TPM_CC_GetRandom": 0x0000017B,
            "TPM_CC_GetTestResult": 0x0000017C,
            "TPM_CC_Hash": 0x0000017D,
            "TPM_CC_PCR_Read": 0x0000017E,
            "TPM_CC_PolicyPCR": 0x0000017F,
            "TPM_CC_PolicyRestart": 0x00000180,
            "TPM_CC_ReadClock": 0x00000181,
            "TPM_CC_PCR_Extend": 0x00000182,
            "TPM_CC_PCR_SetAuthValue": 0x00000183,
            "TPM_CC_NV_Certify": 0x00000184,
            "TPM_CC_EventSequenceComplete": 0x00000185,
            "TPM_CC_HashSequenceStart": 0x00000186,
            "TPM_CC_PolicyPhysicalPresence": 0x00000187,
            "TPM_CC_PolicyDuplicationSelect": 0x00000188,
            "TPM_CC_PolicyGetDigest": 0x00000189,
            "TPM_CC_TestParms": 0x0000018A,
            "TPM_CC_Commit": 0x0000018B,
            "TPM_CC_PolicyPassword": 0x0000018C,
            "TPM_CC_ZGen_2Phase": 0x0000018D,
            "TPM_CC_EC_Ephemeral": 0x0000018E,
        }.get(cc_string, None)

    @staticmethod
    def get_string(cc_code: int) -> Optional[str]:
        """Returns associated command string.

        Args:
            cc_code (int): command code
        Returns:
            (str): Associated command string
            (None): cc_code not found
        """
        return {
            0x0000011F: "TPM_CC_NV_UndefineSpaceSpecial",
            0x00000120: "TPM_CC_EvictControl",
            0x00000121: "TPM_CC_HierarchyControl",
            0x00000122: "TPM_CC_NV_UndefineSpace",
            0x00000124: "TPM_CC_ChangeEPS",
            0x00000125: "TPM_CC_ChangePPS",
            0x00000126: "TPM_CC_Clear",
            0x00000127: "TPM_CC_ClearControl",
            0x00000128: "TPM_CC_ClockSet",
            0x00000129: "TPM_CC_HierarchyChangeAuth",
            0x0000012A: "TPM_CC_NV_DefineSpace",
            0x0000012B: "TPM_CC_PCR_Allocate",
            0x0000012C: "TPM_CC_PCR_SetAuthPolicy",
            0x0000012D: "TPM_CC_PP_Commands",
            0x0000012E: "TPM_CC_SetPrimaryPolicy",
            0x0000012F: "TPM_CC_FieldUpgradeStart",
            0x00000130: "TPM_CC_ClockRateAdjust",
            0x00000131: "TPM_CC_CreatePrimary",
            0x00000132: "TPM_CC_NV_GlobalWriteLock",
            0x00000133: "TPM_CC_GetCommandAuditDigest",
            0x00000134: "TPM_CC_NV_Increment",
            0x00000135: "TPM_CC_NV_SetBits",
            0x00000136: "TPM_CC_NV_Extend",
            0x00000137: "TPM_CC_NV_Write",
            0x00000138: "TPM_CC_NV_WriteLock",
            0x00000139: "TPM_CC_DictionaryAttackLockReset",
            0x0000013A: "TPM_CC_DictionaryAttackParameters",
            0x0000013B: "TPM_CC_NV_ChangeAuth",
            0x0000013C: "TPM_CC_PCR_Event",
            0x0000013D: "TPM_CC_PCR_Reset",
            0x0000013E: "TPM_CC_SequenceComplete",
            0x0000013F: "TPM_CC_SetAlgorithmSet",
            0x00000140: "TPM_CC_SetCommandCodeAuditStatus",
            0x00000141: "TPM_CC_FieldUpgradeData",
            0x00000142: "TPM_CC_IncrementalSelfTest",
            0x00000143: "TPM_CC_SelfTest",
            0x00000144: "TPM_CC_Startup",
            0x00000145: "TPM_CC_Shutdown",
            0x00000146: "TPM_CC_StirRandom",
            0x00000147: "TPM_CC_ActivateCredential",
            0x00000148: "TPM_CC_Certify",
            0x00000149: "TPM_CC_PolicyNV",
            0x0000014A: "TPM_CC_CertifyCreation",
            0x0000014B: "TPM_CC_Duplicate",
            0x0000014C: "TPM_CC_GetTime",
            0x0000014D: "TPM_CC_GetSessionAuditDigest",
            0x0000014E: "TPM_CC_NV_Read",
            0x0000014F: "TPM_CC_NV_ReadLock",
            0x00000150: "TPM_CC_ObjectChangeAuth",
            0x00000151: "TPM_CC_PolicySecret",
            0x00000152: "TPM_CC_Rewrap",
            0x00000153: "TPM_CC_Create",
            0x00000154: "TPM_CC_ECDH_ZGen",
            0x00000155: "TPM_CC_HMAC",
            0x00000156: "TPM_CC_Import",
            0x00000157: "TPM_CC_Load",
            0x00000158: "TPM_CC_Quote",
            0x00000159: "TPM_CC_RSA_Decrypt",
            0x0000015B: "TPM_CC_HMAC_Start",
            0x0000015C: "TPM_CC_SequenceUpdate",
            0x0000015D: "TPM_CC_Sign",
            0x0000015E: "TPM_CC_Unseal",
            0x00000160: "TPM_CC_PolicySigned",
            0x00000161: "TPM_CC_ContextLoad",
            0x00000162: "TPM_CC_ContextSave",
            0x00000163: "TPM_CC_ECDH_KeyGen",
            0x00000164: "TPM_CC_EncryptDecrypt",
            0x00000165: "TPM_CC_FlushContext",
            0x00000167: "TPM_CC_LoadExternal",
            0x00000168: "TPM_CC_MakeCredential",
            0x00000169: "TPM_CC_NV_ReadPublic",
            0x0000016A: "TPM_CC_PolicyAuthorize",
            0x0000016B: "TPM_CC_PolicyAuthValue",
            0x0000016C: "TPM_CC_PolicyCommandCode",
            0x0000016D: "TPM_CC_PolicyCounterTimer",
            0x0000016E: "TPM_CC_PolicyCpHash",
            0x0000016F: "TPM_CC_PolicyLocality",
            0x00000170: "TPM_CC_PolicyNameHash",
            0x00000171: "TPM_CC_PolicyOR",
            0x00000172: "TPM_CC_PolicyTicket",
            0x00000173: "TPM_CC_ReadPublic",
            0x00000174: "TPM_CC_RSA_Encrypt",
            0x00000176: "TPM_CC_StartAuthSession",
            0x00000177: "TPM_CC_VerifySignature",
            0x00000178: "TPM_CC_ECC_Parameters",
            0x00000179: "TPM_CC_FirmwareRead",
            0x0000017A: "TPM_CC_GetCapability",
            0x0000017B: "TPM_CC_GetRandom",
            0x0000017C: "TPM_CC_GetTestResult",
            0x0000017D: "TPM_CC_Hash",
            0x0000017E: "TPM_CC_PCR_Read",
            0x0000017F: "TPM_CC_PolicyPCR",
            0x00000180: "TPM_CC_PolicyRestart",
            0x00000181: "TPM_CC_ReadClock",
            0x00000182: "TPM_CC_PCR_Extend",
            0x00000183: "TPM_CC_PCR_SetAuthValue",
            0x00000184: "TPM_CC_NV_Certify",
            0x00000185: "TPM_CC_EventSequenceComplete",
            0x00000186: "TPM_CC_HashSequenceStart",
            0x00000187: "TPM_CC_PolicyPhysicalPresence",
            0x00000188: "TPM_CC_PolicyDuplicationSelect",
            0x00000189: "TPM_CC_PolicyGetDigest",
            0x0000018A: "TPM_CC_TestParms",
            0x0000018B: "TPM_CC_Commit",
            0x0000018C: "TPM_CC_PolicyPassword",
            0x0000018D: "TPM_CC_ZGen_2Phase",
            0x0000018E: "TPM_CC_EC_Ephemeral",
        }.get(cc_code, None)


class ResponseCode(object):
    """Class representing a Response code.

    Allows alternativing between response codes and response strings.
    """

    @staticmethod
    def get_simple_string(rc_code: int) -> Optional[str]:
        """Returns associated response string.

        Args:
            rc_code: response code
        Returns:
            (str): Associated response string
            (None): rc_code not found
        """
        return {
            0x00000100: "TPM_RC_INITIALIZE",
            0x00000101: "TPM_RC_FAILURE",
            0x00000103: "TPM_RC_SEQUENCE",
            0x0000010B: "TPM_RC_PRIVATE",
            0x00000119: "TPM_RC_HMAC",
            0x00000120: "TPM_RC_DISABLED",
            0x00000121: "TPM_RC_EXCLUSIVE",
            0x00000124: "TPM_RC_AUTH_TYPE",
            0x00000125: "TPM_RC_AUTH_MISSING",
            0x00000126: "TPM_RC_POLICY",
            0x00000127: "TPM_RC_PCR",
            0x00000128: "TPM_RC_PCR_CHANGED",
            0x0000012D: "TPM_RC_UPGRADE",
            0x0000012E: "TPM_RC_TOO_MANY_CONTEXTS",
            0x0000012F: "TPM_RC_AUTH_UNAVAILABLE",
            0x00000130: "TPM_RC_REBOOT",
            0x00000131: "TPM_RC_UNBALANCED",
            0x00000142: "TPM_RC_COMMAND_SIZE",
            0x00000143: "TPM_RC_COMMAND_CODE",
            0x00000144: "TPM_RC_AUTHSIZE",
            0x00000145: "TPM_RC_AUTH_CONTEXT",
            0x00000146: "TPM_RC_NV_RANGE",
            0x00000147: "TPM_RC_NV_SIZE",
            0x00000148: "TPM_RC_NV_LOCKED",
            0x00000149: "TPM_RC_NV_AUTHORIZATION",
            0x0000014A: "TPM_RC_NV_UNINITIALIZED",
            0x0000014B: "TPM_RC_NV_SPACE",
            0x0000014C: "TPM_RC_NV_DEFINED",
            0x00000150: "TPM_RC_BAD_CONTEXT",
            0x00000151: "TPM_RC_CPHASH",
            0x00000152: "TPM_RC_PARENT",
            0x00000153: "TPM_RC_NEEDS_TEST",
            0x00000154: "TPM_RC_NO_RESULT",
            0x00000155: "TPM_RC_SENSITIVE",
        }.get(rc_code, None)

    @staticmethod
    def parse_code(rc_code: int) -> tuple:
        """Returns a Tuple representing the parsed response code."""
        generic_errors = {
            0x000: "TPM_RC_INITIALIZE",
            0x001: "TPM_RC_FAILURE",
            0x003: "TPM_RC_SEQUENCE",
            0x00B: "TPM_RC_PRIVATE",
            0x019: "TPM_RC_HMAC",
            0x020: "TPM_RC_DISABLED",
            0x021: "TPM_RC_EXCLUSIVE",
            0x024: "TPM_RC_AUTH_TYPE",
            0x025: "TPM_RC_AUTH_MISSING",
            0x026: "TPM_RC_POLICY",
            0x027: "TPM_RC_PCR",
            0x028: "TPM_RC_PCR_CHANGED",
            0x02D: "TPM_RC_UPGRADE",
            0x02E: "TPM_RC_TOO_MANY_CONTEXTS",
            0x02F: "TPM_RC_AUTH_UNAVAILABLE",
            0x030: "TPM_RC_REBOOT",
            0x031: "TPM_RC_UNBALANCED",
            0x042: "TPM_RC_COMMAND_SIZE",
            0x043: "TPM_RC_COMMAND_CODE",
            0x044: "TPM_RC_AUTHSIZE",
            0x045: "TPM_RC_AUTH_CONTEXT",
            0x046: "TPM_RC_NV_RANGE",
            0x047: "TPM_RC_NV_SIZE",
            0x048: "TPM_RC_NV_LOCKED",
            0x049: "TPM_RC_NV_AUTHORIZATION",
            0x04A: "TPM_RC_NV_UNINITIALIZED",
            0x04B: "TPM_RC_NV_SPACE",
            0x04C: "TPM_RC_NV_DEFINED",
            0x050: "TPM_RC_BAD_CONTEXT",
            0x051: "TPM_RC_CPHASH",
            0x052: "TPM_RC_PARENT",
            0x053: "TPM_RC_NEEDS_TEST",
            0x054: "TPM_RC_NO_RESULT",
            0x055: "TPM_RC_SENSITIVE",
        }

        handle_errors = {
            0x001: "TPM_RC_ASYMMETRIC",
            0x002: "TPM_RC_ATTRIBUTES",
            0x003: "TPM_RC_HASH",
            0x004: "TPM_RC_VALUE",
            0x005: "TPM_RC_HIERARCHY",
            0x007: "TPM_RC_KEY_SIZE",
            0x008: "TPM_RC_MGF",
            0x009: "TPM_RC_MODE",
            0x00A: "TPM_RC_TYPE",
            0x00B: "TPM_RC_HANDLE",
            0x00C: "TPM_RC_KDF",
            0x00D: "TPM_RC_RANGE",
            0x00E: "TPM_RC_AUTH_FAIL",
            0x00F: "TPM_RC_NONCE",
            0x010: "TPM_RC_PP",
            0x012: "TPM_RC_SCHEME",
            0x015: "TPM_RC_SIZE",
            0x016: "TPM_RC_SYMMETRIC",
            0x017: "TPM_RC_TAG",
            0x018: "TPM_RC_SELECTOR",
            0x01A: "TPM_RC_INSUFFICIENT",
            0x01B: "TPM_RC_SIGNATURE",
            0x01C: "TPM_RC_KEY",
            0x01D: "TPM_RC_POLICY_FAIL",
            0x01F: "TPM_RC_INTEGRITY",
            0x020: "TPM_RC_TICKET",
            0x021: "TPM_RC_RESERVED_BITS",
            0x022: "TPM_RC_BAD_AUTH",
            0x023: "TPM_RC_EXPIRED",
            0x024: "TPM_RC_POLICY_CC",
            0x025: "TPM_RC_BINDING",
            0x026: "TPM_RC_CURVE",
            0x027: "TPM_RC_ECC_POINT",
        }

        warnings = {
            0x001: "TPM_RC_CONTEXT_GAP",
            0x002: "TPM_RC_OBJECT_MEMORY",
            0x003: "TPM_RC_SESSION_MEMORY",
            0x004: "TPM_RC_MEMORY",
            0x005: "TPM_RC_SESSION_HANDLES",
            0x006: "TPM_RC_OBJECT_HANDLES",
            0x007: "TPM_RC_LOCALITY",
            0x008: "TPM_RC_YIELDED",
            0x009: "TPM_RC_CANCELED",
            0x00A: "TPM_RC_TESTING",
            0x010: "TPM_RC_REFERENCE_H0",
            0x011: "TPM_RC_REFERENCE_H1",
            0x012: "TPM_RC_REFERENCE_H2",
            0x013: "TPM_RC_REFERENCE_H3",
            0x014: "TPM_RC_REFERENCE_H4",
            0x015: "TPM_RC_REFERENCE_H5",
            0x016: "TPM_RC_REFERENCE_H6",
            0x018: "TPM_RC_REFERENCE_S0",
            0x019: "TPM_RC_REFERENCE_S1",
            0x01A: "TPM_RC_REFERENCE_S2",
            0x01B: "TPM_RC_REFERENCE_S3",
            0x01C: "TPM_RC_REFERENCE_S4",
            0x01D: "TPM_RC_REFERENCE_S5",
            0x01E: "TPM_RC_REFERENCE_S6",
            0x020: "TPM_RC_NV_RATE",
            0x021: "TPM_RC_LOCKOUT",
            0x022: "TPM_RC_RETRY",
            0x023: "TPM_RC_NV_UNAVAILABLE",
        }

        # Check for TPM_RC_SUCCESS.
        if rc_code == 0x00:
            return ("Success", "None", 0, "TPM_RC_SUCCESS", "NA")

        # Check for TPM 1.2 response.
        if not (rc_code & (0b11 << 7)):
            return ("Tpm1.2 Response", "None", 0, 0, "NA")

        # Check bit 7.
        if not (rc_code & (1 << 7)):
            # Check bit 10.
            if rc_code & (1 << 10):
                return ("Vendor Defined Code", "None", 0, 0, "NA")

            # At this point the code will be in [6:0]...
            code = rc_code & 0b1111111

            # Check bit 11.
            if rc_code & (1 << 11):
                return ("Warning", "None", 0, warnings[code], "NA")  # TODO: Complete this.
            else:
                return ("Error", "None", 0, code, generic_errors[code])

        # At this point the code will always be in [5:0]...
        code = rc_code & 0b111111

        # Check bit 6.
        if rc_code & (1 << 6):
            number = (rc_code >> 8) & 0b1111
            return ("Error", "Parameter", number, code, "NA")  # TODO: Complete this.

        # At this point the nubmer will always be in [10:8]...
        number = (rc_code >> 8) & 0b111

        # Check bit 11.
        if not (rc_code & (1 << 11)):
            return ("Error", "Handle", number, code, handle_errors[code])  # TODO: Complete this.
        else:
            return ("Error", "Session", number, code, "NA")  # TODO: Complete this.

        raise ValueError("Code '0x%x' could not be parsed!" % rc_code)
        return None



##
# Module to generate inf files for capsule update based on INF TEMPLATE and
# supplied information (Name, Version, ESRT Guid, Rollback, etc.)
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module to generate inf files for capsule update based on an INF template.

Uses supplied information such as Name, Version, ESRT Guid, Rollback, etc.
"""

import datetime
import logging
import os
import re
import uuid
from typing import Optional

OS_BUILD_VERSION_DIRID13_SUPPORT = "10.0...17134"


class InfSection(object):
    """Object representing an INF Section."""

    def __init__(self, name: str) -> "InfSection":
        """Inits the object."""
        self.Name = name
        self.Items = []

    def __str__(self) -> str:
        """Returns the string representation of the object.

        Returns:
            (str): string representation of the object.
        """
        return "\n".join(["[%s]" % self.Name] + self.Items)


class InfGenerator(object):
    """An object that generates an INF file from data it is initialized with."""

    ### INF Template ###
    TEMPLATE = r""";
; {Name}.inf
; {DriverVersion}
; Copyright (C) 2019 Microsoft Corporation.  All Rights Reserved.
;
[Version]
Signature="$WINDOWS NT$"
Class=Firmware
ClassGuid={{f2e7dd72-6468-4e36-b6f1-6488f42c1b52}}
Provider=%Provider%
DriverVer={Date},{DriverVersion}
PnpLockdown=1
CatalogFile={Name}.cat

[Manufacturer]
%MfgName% = Firmware,NT{Arch}.{OsBuildVersion}

[Firmware.NT{Arch}.{OsBuildVersion}]
%FirmwareDesc% = Firmware_Install,UEFI\RES_{{{EsrtGuid}}}

[Firmware_Install.NT]
CopyFiles = Firmware_CopyFiles
{Rollback}
{FirmwareCopyFilesSection}

[Firmware_Install.NT.Services]
AddService=,2

[Firmware_Install.NT.Hw]
AddReg = Firmware_AddReg

{FirmwareAddRegSection}

[SourceDisksNames]
1 = %DiskName%

{SourceDisksFilesSection}

[DestinationDirs]
DefaultDestDir = 13

[Strings]
; localizable
Provider     = "{Provider}"
MfgName      = "{MfgName}"
FirmwareDesc = "{Description}"
DiskName     = "Firmware Update"

; non-localizable
DIRID_WINDOWS = 10
REG_DWORD     = 0x00010001
"""

    ROLLBACKTEMPLATE = r"""AddReg    = Firmware_DowngradePolicy_Addreg

;override firmware resource update policy to allow downgrade to lower version
[Firmware_DowngradePolicy_Addreg]
HKLM,SYSTEM\CurrentControlSet\Control\FirmwareResources\{{{EsrtGuid}}},Policy,%REG_DWORD%,1
"""

    SUPPORTED_ARCH = {"amd64": "amd64", "x64": "amd64", "arm": "arm", "arm64": "ARM64", "aarch64": "ARM64"}

    def __init__(
        self,
        name_string: str,
        provider: str,
        esrt_guid: str,
        arch: str,
        description_string: str,
        version_string: str,
        version_hex: str,
    ) -> None:
        """Inits the object with data necessary to generate an INF file.

        Args:
            name_string (str): Name
            provider (str): Provider
            esrt_guid (str): stringified guid, will be converted
            arch (str): architecture
            description_string (str): description
            version_string (str): version
            version_hex (int): version
        """
        self.Name = name_string
        self.Provider = provider
        self.EsrtGuid = esrt_guid
        self.Arch = arch
        self.Description = description_string
        self.VersionString = version_string
        self.VersionHex = version_hex
        self._manufacturer = None  # default for optional feature
        self._date = datetime.date.today()
        self._integrityfile = None

    @property
    def Name(self) -> str:
        """Getter for the Name Attribute."""
        return self._name

    @Name.setter
    def Name(self, value: str) -> None:
        """Setter for the Name Attribute.

        Raises:
            (ValueError): Invalid Characters in name
        """
        # test here for invalid chars
        if not (re.compile(r"[\w-]*$")).match(value):
            logging.critical("Name invalid: '%s'", value)
            raise ValueError("Name has invalid chars.")
        self._name = value

    @property
    def Provider(self) -> str:
        """Getter for the Provider Attribute."""
        return self._provider

    @Provider.setter
    def Provider(self, value: str) -> None:
        """Setter for the Provider Attribute."""
        self._provider = value

    @property
    def Manufacturer(self) -> str:
        """Getter for the Manufacturer Attribute.

        NOTE: Returns Provider attribute if Manufacturer attribute is not set.
        """
        if self._manufacturer is None:
            return self.Provider

        return self._manufacturer

    @Manufacturer.setter
    def Manufacturer(self, value: str) -> None:
        """Setter for the Manufacturer Attribute."""
        self._manufacturer = value

    @property
    def Description(self) -> str:
        """Getter for the Description Attribute."""
        return self._description

    @Description.setter
    def Description(self, value: str) -> None:
        """Setter for the Description Attribute."""
        self._description = value

    @property
    def EsrtGuid(self) -> str:
        """Getter for the EsrtGuid Attribute."""
        return self._esrt_guid

    @EsrtGuid.setter
    def EsrtGuid(self, value: str) -> None:
        """Setter for the EsrtGuid Attribute.

        Raises:
            (Exception): Invalid value
        """
        uuid.UUID(value)  # if this works it is valid...otherwise throws exception
        # todo - make sure it is formatted exactly right
        self._esrt_guid = value

    @property
    def VersionString(self) -> str:
        """Getter for VersionString attribute."""
        return self._versionstring

    @VersionString.setter
    def VersionString(self, value: str) -> None:
        """Setter for the VersionString attribute.

        Raises:
            (ValueError): Invalid format
        """
        c = value.count(".")
        if (c < 1) or (c > 3):
            logging.critical("Version string in invalid format.")
            raise ValueError("VersionString must be in format of xx.xx -> xx.xx.xx.xx")
        self._versionstring = value

    @property
    def VersionHex(self) -> str:
        """Getter for the VersionHex attribute."""
        return "0x%X" % self._versionhex

    @VersionHex.setter
    def VersionHex(self, value: int) -> None:
        """Setter for the VersionHex attribute.

        Raises:
            (ValueError): hex does not fit in a 32but uint
        """
        a = int(value, 0)
        if a > 0xFFFFFFFF:
            logging.critical("VersionHex invalid: '%s'", value)
            raise ValueError("VersionHex must fit within 32bit value range for unsigned integer")
        self._versionhex = a

    @property
    def Arch(self) -> str:
        """Getter for the Arch property."""
        return self._arch

    @Arch.setter
    def Arch(self, value: str) -> None:
        """Setter for the Arch Attribute.

        Raises:
            (ValueError): Unsupported Arch
        """
        key = value.lower()
        if key not in InfGenerator.SUPPORTED_ARCH.keys():
            logging.critical("Arch invalid: '%s'", value)
            raise ValueError("Unsupported Architecture")
        self._arch = InfGenerator.SUPPORTED_ARCH[key]

    @property
    def Date(self) -> str:
        """Getter for the date attribute.

        Formats to a m/d/y str before returning
        """
        return self._date.strftime("%m/%d/%Y")

    @Date.setter
    def Date(self, value: datetime.date) -> None:
        """Setter for the Date attribute.

        Raises:
            (ValueError): not a datetime.date object
        """
        if not isinstance(value, datetime.date):
            raise ValueError("Date must be a datetime.date object")
        self._date = value

    @property
    def IntegrityFilename(self) -> Optional[str]:
        """Getter for the Integrity File Name.

        Transforms value into string.
        """
        return str(self._integrityfile) if self._integrityfile is not None else ""

    @IntegrityFilename.setter
    def IntegrityFilename(self, value: str) -> None:
        """Setter for the IntegrityFile name."""
        self._integrityfile = value

    def MakeInf(self, OutputInfFilePath: os.PathLike, FirmwareBinFileName: str, Rollback: bool = False) -> int:
        """Generates the INF with provided information.

        Args:
            OutputInfFilePath (os.PathLike): Path to existing file
            FirmwareBinFileName (str): File Name
            Rollback (:obj:`bool`, optional): Generate with Rollback template
        """
        RollbackString = ""
        if Rollback:
            RollbackString = InfGenerator.ROLLBACKTEMPLATE.format(EsrtGuid=self.EsrtGuid)

        binfilename = os.path.basename(FirmwareBinFileName)

        copy_files = InfSection("Firmware_CopyFiles")
        copy_files.Items.append(binfilename)

        add_reg = InfSection("Firmware_AddReg")
        add_reg.Items.append("HKR,,FirmwareId,,{{{guid}}}".format(guid=self.EsrtGuid))
        add_reg.Items.append("HKR,,FirmwareVersion,%REG_DWORD%,{version}".format(version=self.VersionHex))
        add_reg.Items.append("HKR,,FirmwareFilename,,%13%\\{file_name}".format(file_name=binfilename))

        disks_files = InfSection("SourceDisksFiles")
        disks_files.Items.append("{file_name} = 1".format(file_name=binfilename))

        if self.IntegrityFilename != "":
            copy_files.Items.append(self.IntegrityFilename)
            add_reg.Items.append(
                "HKR,,FirmwareIntegrityFilename,,%13%\\{file_name}".format(file_name=self.IntegrityFilename)
            )
            disks_files.Items.append("{file_name} = 1".format(file_name=self.IntegrityFilename))

        Content = InfGenerator.TEMPLATE.format(
            Name=self.Name,
            Date=self.Date,
            Arch=self.Arch,
            DriverVersion=self.VersionString,
            EsrtGuid=self.EsrtGuid,
            Provider=self.Provider,
            MfgName=self.Manufacturer,
            OsBuildVersion=OS_BUILD_VERSION_DIRID13_SUPPORT,
            Description=self.Description,
            Rollback=RollbackString,
            FirmwareCopyFilesSection=copy_files,
            FirmwareAddRegSection=add_reg,
            SourceDisksFilesSection=disks_files,
        )

        with open(OutputInfFilePath, "w") as f:
            f.write(Content)

        return 0



##
# File to mark this a python package
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""This package contains parsers for variable EDK2 files and objects."""



# @file
# Module contains helper classes and functions to work with Variable Policy structures
# and substructures.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module containing helper classes and functions to work with Variable Policy structures and substructures."""

import struct
import uuid
from typing import IO

from edk2toollib.uefi.uefi_multi_phase import EfiVariableAttributes


class VariableLockOnVarStatePolicy(object):
    """Object representing the Lock on var state policy c struct.

    typedef struct {
        EFI_GUID Namespace;
        UINT8    Value;
        UINT8    Reserved;
        // CHAR16   Name[];           // Variable Length Field
    } VARIABLE_LOCK_ON_VAR_STATE_POLICY;
    """

    _HdrStructFormat = "<16sBB"
    _HdrStructSize = struct.calcsize(_HdrStructFormat)

    def __init__(self) -> "VariableLockOnVarStatePolicy":
        """Initializes the Variable Lock On Var State Policy."""
        self.Namespace = uuid.UUID(bytes=b"\x00" * 16)
        self.Value = 0
        self.Name = None

    def __str__(self) -> str:
        """String representation of the object."""
        return "VARIABLE_LOCK_ON_VAR_STATE_POLICY(%s, %d, %s)" % (self.Namespace, self.Value, self.Name)

    def decode(self, buffer: IO) -> bytes:
        """Load this object from a bytes buffer.

        Returns:
            (obj): any remaining buffer
        """
        (_namespace, self.Value, _) = struct.unpack(self._HdrStructFormat, buffer[: self._HdrStructSize])

        self.Namespace = uuid.UUID(bytes_le=_namespace)

        # Scan the rest of the buffer for a \x00\x00 to terminate the string.
        buffer = buffer[self._HdrStructSize :]
        if len(buffer) < 4:
            raise ValueError("Buffer too short!")

        string_end = None
        for i in range(0, len(buffer), 2):
            if buffer[i] == 0 and buffer[i + 1] == 0:
                string_end = i + 2
                break

        if string_end is None:
            raise ValueError("String end not detected!")

        self.Name = buffer[:string_end].decode("utf-16").strip("\x00")

        return buffer[string_end:]


class VariablePolicyEntry(object):
    """Object representing the Variable Policy C struct.

    Example:
        typedef struct {
            UINT32   Version;
            UINT16   Size;
            UINT16   OffsetToName;
            EFI_GUID Namespace;
            UINT32   MinSize;
            UINT32   MaxSize;
            UINT32   AttributesMustHave;
            UINT32   AttributesCantHave;
            UINT8    LockPolicyType;
            UINT8    Reserved[3];
            // UINT8    LockPolicy[];     // Variable Length Field
            // CHAR16   Name[]            // Variable Length Field
        } VARIABLE_POLICY_ENTRY;
    """

    _HdrStructFormat = "<IHH16sIIIIB3s"  # spell-checker:disable-line
    _HdrStructSize = struct.calcsize(_HdrStructFormat)

    ENTRY_REVISION = 0x0001_0000

    NO_MIN_SIZE = 0
    NO_MAX_SIZE = 0xFFFF_FFFF
    NO_MUST_ATTR = 0
    NO_CANT_ATTR = 0

    TYPE_NO_LOCK = 0
    TYPE_LOCK_NOW = 1
    TYPE_LOCK_ON_CREATE = 2
    TYPE_LOCK_ON_VAR_STATE = 3

    LOCK_POLICY_STRING_MAP = {
        TYPE_NO_LOCK: "NONE",
        TYPE_LOCK_NOW: "NOW",
        TYPE_LOCK_ON_CREATE: "ON_CREATE",
        TYPE_LOCK_ON_VAR_STATE: "ON_VAR_STATE",
    }

    def __init__(self) -> "VariablePolicyEntry":
        """Initializes the Variable Policy Entry."""
        self.Version = VariablePolicyEntry.ENTRY_REVISION
        self.Size = VariablePolicyEntry._HdrStructSize
        self.OffsetToName = self.Size
        self.Namespace = uuid.UUID(bytes=b"\x00" * 16)
        self.MinSize = VariablePolicyEntry.NO_MIN_SIZE
        self.MaxSize = VariablePolicyEntry.NO_MAX_SIZE
        self.AttributesMustHave = VariablePolicyEntry.NO_MUST_ATTR
        self.AttributesCantHave = VariablePolicyEntry.NO_CANT_ATTR
        self.LockPolicyType = VariablePolicyEntry.TYPE_NO_LOCK
        self.LockPolicy = None
        self.Name = None

    def __str__(self) -> str:
        """String representation of the object."""
        result = "VARIABLE_POLICY_ENTRY(%s, %s)\n" % (self.Namespace, self.Name)

        if self.LockPolicyType in (
            VariablePolicyEntry.TYPE_NO_LOCK,
            VariablePolicyEntry.TYPE_LOCK_NOW,
            VariablePolicyEntry.TYPE_LOCK_ON_CREATE,
        ):
            result += "\tLock        = %s\n" % VariablePolicyEntry.LOCK_POLICY_STRING_MAP[self.LockPolicyType]
        elif self.LockPolicyType is VariablePolicyEntry.TYPE_LOCK_ON_VAR_STATE:
            result += "\tLock        = %s\n" % self.LockPolicy

        result += "\tMin = 0x%08X, Max = 0x%08X, Must = 0x%08X, Cant = 0x%08X\n" % (
            self.MinSize,
            self.MaxSize,
            self.AttributesMustHave,
            self.AttributesCantHave,
        )

        return result

    @staticmethod
    def csv_header() -> list:
        """Returns a list containing the names of the ordered columns that are produced by csv_row()."""
        return [
            "Namespace",
            "Name",
            "LockPolicyType",
            "VarStateNamespace",
            "VarStateName",
            "VarStateValue",
            "MinSize",
            "MaxSize",
            "AttributesMustHave",
            "AttributesCantHave",
        ]

    def csv_row(self, guid_xref: dict = None) -> list:
        """Returns a list containing the elements of this structure.

        (in the same order as the csv_header) ready to be written to a csv file
        Args:
            guid_xref (dict): GUID/name substitutions where the key is a uuid object
                and the value is a string
        """
        if guid_xref is None:
            guid_xref = {}

        result = [
            guid_xref.get(self.Namespace, self.Namespace),
            self.Name,
            VariablePolicyEntry.LOCK_POLICY_STRING_MAP[self.LockPolicyType],
        ]

        if self.LockPolicyType in (
            VariablePolicyEntry.TYPE_NO_LOCK,
            VariablePolicyEntry.TYPE_LOCK_NOW,
            VariablePolicyEntry.TYPE_LOCK_ON_CREATE,
        ):
            result += ["N/A", "N/A", "N/A"]
        elif self.LockPolicyType is VariablePolicyEntry.TYPE_LOCK_ON_VAR_STATE:
            result += [
                guid_xref.get(self.LockPolicy.Namespace, self.LockPolicy.Namespace),
                self.LockPolicy.Name,
                self.LockPolicy.Value,
            ]

        result += [
            "0x%08X" % self.MinSize,
            "0x%08X" % self.MaxSize,
            str(EfiVariableAttributes(self.AttributesMustHave)),
            str(EfiVariableAttributes(self.AttributesCantHave)),
        ]

        return result

    def decode(self, buffer: IO) -> bytes:
        """Load this object from a bytes buffer.

        Returns:
            (bytes): Any remaining buffer
        """
        (
            self.Version,
            self.Size,
            self.OffsetToName,
            _namespace,
            self.MinSize,
            self.MaxSize,
            self.AttributesMustHave,
            self.AttributesCantHave,
            self.LockPolicyType,
            _,
        ) = struct.unpack(self._HdrStructFormat, buffer[: self._HdrStructSize])

        if self.Version != VariablePolicyEntry.ENTRY_REVISION:
            raise ValueError("Unknown structure version!")
        if self.LockPolicyType not in VariablePolicyEntry.LOCK_POLICY_STRING_MAP:
            raise ValueError("Unknown LockPolicyType!")

        self.Namespace = uuid.UUID(bytes_le=_namespace)

        if self.OffsetToName != self.Size:
            self.Name = buffer[self.OffsetToName : self.Size].decode("utf-16").strip("\x00")

        if self.LockPolicyType == VariablePolicyEntry.TYPE_LOCK_ON_VAR_STATE:
            self.LockPolicy = VariableLockOnVarStatePolicy()
            self.LockPolicy.decode(buffer[self._HdrStructSize : self.OffsetToName])

        return buffer[self.Size :]



# @file
# Contains classes and helper functions to modify variables in a UEFI ROM image.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Contains classes and helper functions to modify variables in a UEFI ROM image."""

import mmap
import os
from typing import Optional

import edk2toollib.uefi.edk2.variable_format as VF
import edk2toollib.uefi.pi_firmware_volume as PiFV


class VariableStore(object):
    """Class representing the variable store."""

    def __init__(
        self, romfile: str, store_base: Optional[int] = None, store_size: Optional[int] = None
    ) -> "VariableStore":
        """Initialize the Variable store and read necessary files.

        Loads the data.
        """
        self.rom_file_path = romfile
        self.store_base = store_base
        self.store_size = store_size
        self.rom_file = None
        self.rom_file_map = None

        if not os.path.isfile(self.rom_file_path):
            raise Exception("'%s' is not the path to a file!" % self.rom_file_path)

        self.rom_file = open(self.rom_file_path, "r+b")
        self.rom_file_map = mmap.mmap(self.rom_file.fileno(), 0)

        # Sanity check some things.
        file_size = self.rom_file_map.size()
        if store_base is not None and store_size is not None and (store_base + store_size) > file_size:
            raise Exception("ROM file is %d bytes. Cannot seek to %d+%d bytes!" % (file_size, store_base, store_size))

        # Go ahead and advance the file cursor and load the FV header.
        self.rom_file.seek(self.store_base)
        self.fv_header = PiFV.EfiFirmwareVolumeHeader().load_from_file(self.rom_file)
        if self.fv_header.FileSystemGuid != PiFV.EfiSystemNvDataFvGuid:
            raise Exception("Store_base is not pointing at a valid SystemNvData FV!")
        if self.fv_header.FvLength != self.store_size:
            raise Exception("Store_size %d does not match FV size %d!" % (self.store_size, self.fv_header.FvLength))

        # Advance the file cursor and load the VarStore header.
        self.rom_file.seek(self.fv_header.HeaderLength, os.SEEK_CUR)
        self.var_store_header = VF.VariableStoreHeader().load_from_file(self.rom_file)
        if (
            self.var_store_header.Format != VF.VARIABLE_STORE_FORMATTED
            or self.var_store_header.State != VF.VARIABLE_STORE_HEALTHY
        ):
            raise Exception("VarStore is invalid or cannot be processed with this helper!")

        # Now we're finally ready to read some variables.
        self.variables = []
        self.rom_file.seek(self.var_store_header.StructSize, os.SEEK_CUR)
        try:
            while True:
                new_var = self.get_new_var_class().load_from_file(self.rom_file)

                # Seek past the current variable in the store.
                self.rom_file.seek(new_var.get_buffer_size(), os.SEEK_CUR)

                # Add the variable to the array.
                self.variables.append(new_var)
        except EOFError:
            pass
        except:
            raise

        # Finally, reset the file cursor to the beginning of the VarStore FV.
        self.rom_file.seek(self.store_base)

    def __del__(self) -> None:
        """Flushes and closes files."""
        if self.rom_file_map is not None:
            self.rom_file_map.flush()
            self.rom_file_map.close()

        if self.rom_file is not None:
            self.rom_file.close()

    def get_new_var_class(self) -> VF.VariableHeader | VF.AuthenticatedVariableHeader:
        """Var class builder method depending on var type."""
        if self.var_store_header.Type == "Var":
            new_var = VF.VariableHeader()
        else:
            new_var = VF.AuthenticatedVariableHeader()

        return new_var

    def add_variable(self, new_var: VF.VariableHeader | VF.AuthenticatedVariableHeader) -> None:
        """Add a variable to the variable list."""
        self.variables.append(new_var)

    def flush_to_file(self) -> None:
        """Flush the changes to file."""
        # First, we need to make sure that our variables will fit in the VarStore.
        var_size = sum([var.get_buffer_size() for var in self.variables])
        # Add the terminating var header.
        dummy_var = self.get_new_var_class()
        var_size += dummy_var.StructSize
        if var_size > self.var_store_header.Size:
            raise Exception(
                "Total variable size %d is too large to fit in VarStore %d!" % (var_size, self.var_store_header.Size)
            )

        # Now, we just have to serialize each variable in turn and write them to the mmap buffer.
        var_offset = self.store_base + self.fv_header.HeaderLength + self.var_store_header.StructSize
        for var in self.variables:
            var_buffer_size = var.get_buffer_size()
            self.rom_file_map[var_offset : (var_offset + var_buffer_size)] = var.serialize(True)
            var_offset += var_buffer_size

        # Add a terminating Variable Header.
        self.rom_file_map[var_offset : (var_offset + dummy_var.StructSize)] = b"\xff" * dummy_var.StructSize

        # Now we have to flush the mmap to the file.
        self.rom_file_map.flush()



## @file
# Module that encodes and decodes a EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER with
# a payload.
#
# Copyright (c) 2018 - 2019, Intel Corporation. All rights reserved.<BR>
# SPDX-License-Identifier: BSD-2-Clause-Patent
#

"""Module for encoding and decoding EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER with payloads."""

import struct
import uuid
from typing import IO

from edk2toollib.uefi.fmp_auth_header import FmpAuthHeaderClass


class FmpCapsuleImageHeaderClass(object):
    r"""An object representing an EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER.

    Can parse or produce an EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER structure/byte buffer.

    Attributes:
        Version (int):                 EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER_INIT_VERSION
        UpdateImageTypeId (uuid.UUID): Used to identify device firmware targeted by this update. This guid is matched
                                       by system firmware against ImageTypeId field within a
                                       EFI_FIRMWARE_IMAGE_DESCRIPTOR
        UpdateImageIndex (int):        Passed as ImageIndex in call to EFI_FIRMWARE_MANAGEMENT_PROTOCOL.SetImage ()
        UpdateImageSize (int):         Size of the binary update image which immediately follows this structure
        UpdateVendorCodeSize (int):    Size of the VendorCode bytes which optionally immediately follow binary update
                                       image in the capsule
        UpdateHardwareInstance (int):  The HardwareInstance to target with this update. If value is zero it means match
                                       all HardwareInstances.
        Payload (str):                 String representing payload as bytes (i.e. b'\x01\x00\x03')

    ```
    typedef struct {
        UINT32   Version;
        EFI_GUID UpdateImageTypeId;
        UINT8    UpdateImageIndex;
        UINT8    reserved_bytes[3];
        UINT32   UpdateImageSize;
        UINT32   UpdateVendorCodeSize;
        UINT64   UpdateHardwareInstance;
    } EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER;

    #define EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER_INIT_VERSION 0x00000002
    ```
    """

    _StructFormat = "<I16sB3BIIQ"  # spell-checker: disable-line
    _StructSize = struct.calcsize(_StructFormat)

    EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER_INIT_VERSION = 0x00000002

    def __init__(self) -> "FmpCapsuleImageHeaderClass":
        """Inits an empty object."""
        self.Version = self.EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER_INIT_VERSION
        self.UpdateImageTypeId = uuid.UUID("00000000-0000-0000-0000-000000000000")
        self.UpdateImageIndex = 0
        self.UpdateImageSize = 0
        self.UpdateVendorCodeSize = 0
        self.UpdateHardwareInstance = 0x0000000000000000
        self.Payload = b""
        self.VendorCodeBytes = b""
        self.FmpAuthHeader = None

    def Encode(self) -> bytes:
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        # If we have an FmpAuthHeader, let's collapse that now.
        if self.FmpAuthHeader is not None:
            self.Payload = self.FmpAuthHeader.Encode()

        self.UpdateImageSize = len(self.Payload)
        self.UpdateVendorCodeSize = len(self.VendorCodeBytes)
        FmpCapsuleImageHeader = struct.pack(
            self._StructFormat,
            self.Version,
            self.UpdateImageTypeId.bytes_le,
            self.UpdateImageIndex,
            0,
            0,
            0,
            self.UpdateImageSize,
            self.UpdateVendorCodeSize,
            self.UpdateHardwareInstance,
        )
        return FmpCapsuleImageHeader + self.Payload + self.VendorCodeBytes

    def Decode(self, Buffer: IO) -> bytes:
        """Loads data into the object from a filestream.

        Args:
            Buffer: Buffer containing data

        Returns:
            bytes: remaining buffer

        Raises:
            (ValueError): Invalid buffer length
            (ValueError): Invalid Version
        """
        if len(Buffer) < self._StructSize:
            raise ValueError
        (
            Version,
            UpdateImageTypeId,
            UpdateImageIndex,
            r0,
            r1,
            r2,
            UpdateImageSize,
            UpdateVendorCodeSize,
            UpdateHardwareInstance,
        ) = struct.unpack(self._StructFormat, Buffer[0 : self._StructSize])

        if Version < self.EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER_INIT_VERSION:
            raise ValueError
        if UpdateImageIndex < 1:
            raise ValueError
        if UpdateImageSize + UpdateVendorCodeSize != len(Buffer[self._StructSize :]):
            raise ValueError

        self.Version = Version
        self.UpdateImageTypeId = uuid.UUID(bytes_le=UpdateImageTypeId)
        self.UpdateImageIndex = UpdateImageIndex
        self.UpdateImageSize = UpdateImageSize
        self.UpdateVendorCodeSize = UpdateVendorCodeSize
        self.UpdateHardwareInstance = UpdateHardwareInstance
        self.Payload = Buffer[self._StructSize : self._StructSize + UpdateImageSize]
        if len(self.Payload) > 0:
            self.FmpAuthHeader = FmpAuthHeaderClass()
            self.FmpAuthHeader.Decode(self.Payload)
        self.VendorCodeBytes = Buffer[self._StructSize + UpdateImageSize :]
        return Buffer[self._StructSize :]

    def DumpInfo(self) -> None:
        """Prints object to Console."""
        print(
            "EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER.Version                = {Version:08X}".format(
                Version=self.Version
            )
        )
        print(
            "EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER.UpdateImageTypeId      = {UpdateImageTypeId}".format(
                UpdateImageTypeId=str(self.UpdateImageTypeId).upper()
            )
        )
        print(
            "EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER.UpdateImageIndex       = {UpdateImageIndex:08X}".format(
                UpdateImageIndex=self.UpdateImageIndex
            )
        )
        print(
            "EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER.UpdateImageSize        = {UpdateImageSize:08X}".format(
                UpdateImageSize=self.UpdateImageSize
            )
        )
        print(
            "EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER.UpdateVendorCodeSize   = {UpdateVendorCodeSize:08X}".format(
                UpdateVendorCodeSize=self.UpdateVendorCodeSize
            )
        )
        print(
            "EFI_FIRMWARE_MANAGEMENT_CAPSULE_IMAGE_HEADER.UpdateHardwareInstance = {UpdateHardwareInstance:016X}".format(
                UpdateHardwareInstance=self.UpdateHardwareInstance
            )
        )
        print(
            "sizeof (Payload)                                                    = {Size:08X}".format(
                Size=len(self.Payload)
            )
        )
        print(
            "sizeof (VendorCodeBytes)                                            = {Size:08X}".format(
                Size=len(self.VendorCodeBytes)
            )
        )
        if self.FmpAuthHeader is not None:
            self.FmpAuthHeader.DumpInfo()


class FmpCapsuleHeaderClass(object):
    """An object representing a EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER_INIT_VERSION.

    Can parse or produce an EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER_INIT_VERSION structure/byte buffer.

    Attributes:
        Version (int): EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER_INIT_VERSION
        EmbeddedDriverCount (int): The number of drivers included in the capsule and the number of corresponding
                                   offsets stored in ItemOffsetList array.
        PayloadItemCount (int):    The number of payload items included in the capsule and the number of corresponding
                                   offsets stored in the ItemOffsetList array.

    ```
    typedef struct {
        UINT32 Version;
        UINT16 EmbeddedDriverCount;
        UINT16 PayloadItemCount;
        UINT64 ItemOffsetList[];
    } EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER;
    #define EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER_INIT_VERSION       0x00000001
    ```
    """

    _StructFormat = "<IHH"
    _StructSize = struct.calcsize(_StructFormat)

    _ItemOffsetFormat = "<Q"
    _ItemOffsetSize = struct.calcsize(_ItemOffsetFormat)

    EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER_INIT_VERSION = 0x00000001

    def __init__(self) -> "FmpCapsuleHeaderClass":
        """Inits an empty object."""
        self.Version = self.EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER_INIT_VERSION
        self.EmbeddedDriverCount = 0
        self.PayloadItemCount = 0
        self._EmbeddedDriverList = []
        self._FmpCapsuleImageHeaderList = []

    def AddEmbeddedDriver(self, EmbeddedDriver: bytes) -> None:
        """Adds an embedded driver to the list."""
        self._EmbeddedDriverList.append(EmbeddedDriver)
        self.EmbeddedDriverCount += 1

    def GetEmbeddedDriver(self, Index: int) -> bytes:
        """Returns the embedded driver at the index."""
        return self._EmbeddedDriverList[Index]

    def AddFmpCapsuleImageHeader(self, FmpCapsuleHeader: "FmpCapsuleImageHeaderClass") -> None:
        """Adds an Fmp Capsule Image header to the list."""
        self._FmpCapsuleImageHeaderList.append(FmpCapsuleHeader)
        self.PayloadItemCount += 1

    def GetFmpCapsuleImageHeader(self, Index: int) -> "FmpCapsuleImageHeaderClass":
        """Returns the Fmp Capsule Image header at the index."""
        return self._FmpCapsuleImageHeaderList[Index]

    def Encode(self) -> bytes:
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        self.EmbeddedDriverCount = len(self._EmbeddedDriverList)
        self.PayloadItemCount = len(self._FmpCapsuleImageHeaderList)

        FmpCapsuleHeader = struct.pack(
            self._StructFormat, self.Version, self.EmbeddedDriverCount, self.PayloadItemCount
        )

        FmpCapsuleData = b""
        offset_list = []
        Offset = self._StructSize + (self.EmbeddedDriverCount + self.PayloadItemCount) * self._ItemOffsetSize
        for EmbeddedDriver in self._EmbeddedDriverList:
            FmpCapsuleData = FmpCapsuleData + EmbeddedDriver
            offset_list.append(Offset)
            Offset = Offset + len(EmbeddedDriver)
        for FmpCapsuleImageHeader in self._FmpCapsuleImageHeaderList:
            FmpCapsuleImage = FmpCapsuleImageHeader.Encode()
            FmpCapsuleData = FmpCapsuleData + FmpCapsuleImage
            offset_list.append(Offset)
            Offset = Offset + len(FmpCapsuleImage)

        for Offset in offset_list:
            FmpCapsuleHeader = FmpCapsuleHeader + struct.pack(self._ItemOffsetFormat, Offset)

        return FmpCapsuleHeader + FmpCapsuleData

    def Decode(self, Buffer: IO) -> bytes:
        """Loads data into the object from a Buffer.

        Args:
            Buffer (obj): The buffer containing the data.

        Returns:
            (obj): Reaming buffer

        Raises:
            (ValueError): Invalid Buffer
            (ValueError): Invalid Version
        """
        if len(Buffer) < self._StructSize:
            raise ValueError
        (Version, EmbeddedDriverCount, PayloadItemCount) = struct.unpack(
            self._StructFormat, Buffer[0 : self._StructSize]
        )
        if Version < self.EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER_INIT_VERSION:
            raise ValueError

        self.Version = Version
        self.EmbeddedDriverCount = EmbeddedDriverCount
        self.PayloadItemCount = PayloadItemCount
        self._EmbeddedDriverList = []
        self._FmpCapsuleImageHeaderList = []

        offset_list = []

        #
        # Parse the ItemOffsetList values
        #
        Offset = self._StructSize
        for Index in range(EmbeddedDriverCount + PayloadItemCount):
            ItemOffset = struct.unpack(self._ItemOffsetFormat, Buffer[Offset : Offset + self._ItemOffsetSize])[0]
            if ItemOffset >= len(Buffer):
                raise ValueError
            offset_list.append(ItemOffset)
            Offset = Offset + self._ItemOffsetSize
        Result = Buffer[Offset:]

        #
        # Parse the EmbeddedDrivers
        #
        for Index in range(EmbeddedDriverCount):
            Offset = offset_list[Index]
            if Index < (len(offset_list) - 1):
                Length = offset_list[Index + 1] - Offset
            else:
                Length = len(Buffer) - Offset
            self.AddEmbeddedDriver(Buffer[Offset : Offset + Length])

        #
        # Parse the Payloads that are FMP Capsule Images
        #
        for Index in range(EmbeddedDriverCount, EmbeddedDriverCount + PayloadItemCount):
            Offset = offset_list[Index]
            if Index < (len(offset_list) - 1):
                Length = offset_list[Index + 1] - Offset
            else:
                Length = len(Buffer) - Offset
            FmpCapsuleImageHeader = FmpCapsuleImageHeaderClass()
            FmpCapsuleImageHeader.Decode(Buffer[Offset : Offset + Length])
            self.AddFmpCapsuleImageHeader(FmpCapsuleImageHeader)

        return Result

    def DumpInfo(self) -> None:
        """Prints the object to the console."""
        print("EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER.Version             = {Version:08X}".format(Version=self.Version))
        print(
            "EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER.EmbeddedDriverCount = {EmbeddedDriverCount:08X}".format(
                EmbeddedDriverCount=self.EmbeddedDriverCount
            )
        )
        for EmbeddedDriver in self._EmbeddedDriverList:
            print(
                "  sizeof (EmbeddedDriver)                                  = {Size:08X}".format(
                    Size=len(EmbeddedDriver)
                )
            )
        print(
            "EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER.PayloadItemCount    = {PayloadItemCount:08X}".format(
                PayloadItemCount=self.PayloadItemCount
            )
        )
        print("EFI_FIRMWARE_MANAGEMENT_CAPSULE_HEADER.ItemOffsetList      = ")
        for FmpCapsuleImageHeader in self._FmpCapsuleImageHeaderList:
            FmpCapsuleImageHeader.DumpInfo()



# @file tpm2_simulator.py
# This file contains transportation layer classes for interacting with the TPM 2.0 simulator.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module that contains transportation layer classes for interacting with the TPM 2.0 simulator."""

import socket
import struct

import edk2toollib.tpm.tpm2_defs as t2d
import edk2toollib.tpm.tpm2_stream as t2s

PLAT_COMMANDS = {
    "TPM_SIGNAL_POWER_ON": 1,
    "TPM_SIGNAL_POWER_OFF": 2,
    "TPM_SIGNAL_PHYS_PRES_ON": 3,
    "TPM_SIGNAL_PHYS_PRES_OFF": 4,
    "TPM_SIGNAL_HASH_START": 5,
    "TPM_SIGNAL_HASH_DATA": 6,
    # {UINT32 BufferSize, BYTE[BufferSize] Buffer}
    "TPM_SIGNAL_HASH_END": 7,
    "TPM_SEND_COMMAND": 8,
    # {BYTE Locality, UINT32 InBufferSize, BYTE[InBufferSize] InBuffer} ->
    # {UINT32 OutBufferSize, BYTE[OutBufferSize] OutBuffer}
    "TPM_SIGNAL_CANCEL_ON": 9,
    "TPM_SIGNAL_CANCEL_OFF": 10,
    "TPM_SIGNAL_NV_ON": 11,
    "TPM_SIGNAL_NV_OFF": 12,
    "TPM_SIGNAL_KEY_CACHE_ON": 13,
    "TPM_SIGNAL_KEY_CACHE_OFF": 14,
    "TPM_REMOTE_HANDSHAKE": 15,
    "TPM_SET_ALTERNATIVE_RESULT": 16,
    "TPM_SIGNAL_RESET": 17,
    "TPM_SESSION_END": 20,
    "TPM_STOP": 21,
    "TPM_GET_COMMAND_RESPONSE_SIZES": 25,
    "TPM_TEST_FAILURE_MODE": 30,
}


class TpmSimulator(object):
    """An object for interacting with the Tpm Simulator."""

    def __init__(self, host: str = "localhost", port: int = 2321) -> "TpmSimulator":
        """Initialize the simulator on the requested host (ip) and port."""
        super(TpmSimulator, self).__init__()

        # Connect to the control socket.
        self.platSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.platSock.connect((host, port + 1))

        # Connect to the simulator socket.
        self.tpmSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.tpmSock.connect((host, port))

        # Power cycle the TPM.
        self.platSock.send(struct.pack(">L", PLAT_COMMANDS["TPM_SIGNAL_POWER_OFF"]))
        self.platSock.send(struct.pack(">L", PLAT_COMMANDS["TPM_SIGNAL_POWER_ON"]))

        # Enable the NV space.
        self.platSock.send(struct.pack(">L", PLAT_COMMANDS["TPM_SIGNAL_NV_ON"]))

    def send_raw_data(self, data: str) -> None:
        """Send raw data to the TPM simulator."""
        print("RAW -->: " + str(data).encode("hex"))
        self.tpmSock.send(data)

    def read_raw_data(self, count: int) -> bytes:
        """Read raw data from the TPM simulator."""
        data = self.tpmSock.recv(count)
        print("RAW <--: " + str(data).encode("hex"))
        return data

    def send_data(self, data: str) -> bytes:
        """Send data to the TPM simulator."""
        # Send the "I'm about to send data" command.
        self.send_raw_data(struct.pack(">L", PLAT_COMMANDS["TPM_SEND_COMMAND"]))
        # Send the locality for the data.
        self.send_raw_data(struct.pack(">b", 0x03))
        # Send the size of the data.
        self.send_raw_data(struct.pack(">L", len(data)))

        # Now, send the data itself.
        self.send_raw_data(data)

        # Poll until a result is available.
        # NOTE: This shouldn't be necessary and denotes a lack of understanding...
        while True:
            result_size = self.read_raw_data(4)
            result_size = struct.unpack(">L", result_size)[0]
            if result_size > 0:
                break

        return self.read_raw_data(result_size)

    def startup(self, type: str) -> bytes:
        """Initialize the connection to the TPM simulator."""
        stream = t2s.Tpm2CommandStream(t2d.TPM_ST_NO_SESSIONS, 0x00, t2d.TPM_CC_Startup)
        stream.add_element(t2s.Tpm2StreamPrimitive(t2d.TPM_SU_Size, type))
        return self.send_data(stream.get_stream())



# @file tpm2_stream.py
# This file contains utility classes to help marshal and un-marshal data to/from the TPM.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module that contains utility classes to help marshal and un-marshal date to/from the TPM."""

import struct


class Tpm2StreamElement(object):
    """Tpm2 Stream Element."""

    def __init__(self) -> "Tpm2StreamElement":
        """Init an empty Tpm2StreamElement."""
        self.pack_string = ""

    def get_size(self) -> int:
        """The size of this structure when marshalled."""
        return struct.calcsize(self.pack_string)


class Tpm2StreamPrimitive(Tpm2StreamElement):
    """Tpm2 Stream Primitive.

    Attributes:
        size: size of the primitive. 1, 2, 4, or 8 bytes
        value: Value of primitive
    """

    def __init__(self, size: int, value: str) -> "Tpm2StreamPrimitive":
        """Init a primitive value.

        Args:
            size: 1, 2, 4, or 8 bytes
            value: Value to stream.
        """
        super(Tpm2StreamPrimitive, self).__init__()

        if size not in (1, 2, 4, 8):
            raise ValueError("Size must be 1, 2, 4, or 8 bytes!")

        self.pack_string = {1: ">B", 2: ">H", 4: ">L", 8: ">Q"}[size]
        self.value = value

    def marshal(self) -> bytes:
        r"""Serializes the Tpm2 primitive.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        return struct.pack(self.pack_string, self.value)


class TPM2_COMMAND_HEADER(Tpm2StreamElement):
    """Tpm2 Command header.

    Attributes:
        tag: The Tag
        code: The Code
        size: The size of the code
    """

    def __init__(self, tag: str, size: str, code: str) -> "TPM2_COMMAND_HEADER":
        """Init a Tpm2 command."""
        super(TPM2_COMMAND_HEADER, self).__init__()
        self.tag = tag
        self.code = code
        self.size = size
        self.pack_string = ">HLL"

    def update_size(self, size: int) -> None:
        """Update size of the whole command."""
        self.size = size

    def marshal(self) -> str:
        r"""Serializes the Tpm2 command header.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        return struct.pack(self.pack_string, self.tag, self.size, self.code)


class TPM2B(Tpm2StreamElement):
    """Tpm2 B."""

    def __init__(self, data: str) -> "TPM2B":
        """Inits the object."""
        super(TPM2B, self).__init__()
        self.data = data
        self.size = len(data)
        self.pack_string = ">H%ds" % self.size

    def update_data(self, data: str) -> None:
        """Updates the data attribute."""
        self.data = data
        self.size = len(data)
        self.pack_string = ">H%ds" % self.size

    def marshal(self) -> str:
        r"""Serializes the Tpm2B object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        return struct.pack(self.pack_string, self.size, self.data)


class Tpm2CommandStream(object):
    """Tpm2 Command Stream."""

    def __init__(self, tag: str, size: int, code: str) -> "Tpm2CommandStream":
        """Inits a Tpm2 Command stream object."""
        super(Tpm2CommandStream, self).__init__()
        self.header = TPM2_COMMAND_HEADER(tag, size, code)
        self.stream_size = self.header.get_size()
        self.header.update_size(self.stream_size)
        self.stream_elements = []

    def get_size(self) -> int:
        """Returns the stream size."""
        return self.stream_size

    def add_element(self, element: "Tpm2StreamElement") -> None:
        """Adds an element to the stream list."""
        self.stream_elements.append(element)
        self.stream_size += element.get_size()
        self.header.update_size(self.stream_size)

    def get_stream(self) -> str:
        r"""Serializes the Header + elements.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        return self.header.marshal() + b"".join(element.marshal() for element in self.stream_elements)



# @file
#
# Library to parse/manage/build unsigned firmware policy blobs
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

# spell-checker:ignore QWORD

"""A Firmware policy management tools.

Summary:
A firmware policy is conceptually a set of key value pair "rules".
Rules can be 1-time actions, states persistent while the policy is in effect,
or targeting information describing the machine where the policy is applicable

Each key is composed of a UINT32 RootKey followed by String SubKeyName & ValueName
For example:
  RootKey = 0xEF100000
  SubKeyName = "PolicyGroupFoo"
  ValueName = "TpmClear"
  MyExampleKey = EF100000_PolicyGroupFoo_TpmClear

Each value has a type that can be primitive or a variable-length structure.

##
The firmware policy binary, pack(1), little endian, is as follows:

UINT16   FormatVersion
UINT32   PolicyVersion
GUID     PolicyPublisher
UINT16   Reserved1Count  # 0
         Reserved1[Reserved1Count]   # not present, not supported by this library
UINT32   OptionFlags     # 0
UINT16   Reserved2Count  # 0
UINT16   RulesCount
         Reserved2[Reserved2Count]   # not present for FW policies, partial support by this library
RULE     Rules[RulesCount]  # inline, not a pointer, see class Rule below
BYTE     ValueTable[]  #inline, not a pointer
##

##
A RULE structure is as follows

UINT32 RootKey
UINT32 OffsetToSubKeyName  # ValueTable offset
UINT32 OffsetToValueName   # ValueTable offset
UINT32 OffsetToValue       # ValueTable offset to a PolicyValueType + PolicyValue
##

##
Each Rule corresponds to 3 entries in the ValueTable
Rules may, however, re-use identical ValueTable entries to save space
For example, multiple Rules with the same SubKeyName may report the same SubKeyNameOffset

PolicyString SubKeyName  # may or may not be NULL terminated, usually not
PolicyString Valuename   # may or may not be NULL terminated, usually not
PolicyValue              # Begins with a UINT16 PolicyValueType, followed by the actual value, may be NULL terminated
##

##
Conventions

    _init__ constructors are to be used when creating new objects from desired values.
    When the number of parameters is small, __init__ may also take a stream parameter
    which will be deserialized to initialize the object in lieu of the other parameters.
    When the number of parameters is large, initialization from stream is factored out
    into a separate class method FromFileStream().  This overhead did not seem merited
    when the number of overall parameters was small.
##
"""

import io
import struct
import uuid
from typing import BinaryIO

STRICT = False


class Rule(object):
    """Class for managing Rule structures.

    Class for storing, serializing, deserializing, & printing RULE elements
    """

    StructFormat = "<IIII"
    StructSize = struct.calcsize(StructFormat)

    def __init__(
        self,
        RootKey: int,
        SubKeyName: str,
        ValueName: str,
        Value: "PolicyValue",
        OffsetToSubKeyName: int = 0,
        OffsetToValueName: int = 0,
        OffsetToValue: int = 0,
    ) -> "Rule":
        """Inits the Rule Object.

        Args:
            RootKey (int): The root key value
            SubKeyName (str): the subkey name
            ValueName (str): Name of the rule value
            Value (PolicyValue): The value.
            OffsetToSubKeyName (:obj:`int`, optional): Offset to read the subkey name
            OffsetToValueName (:obj:`int`, optional): Offset to read the value name
            OffsetToValue (:obj:`int`, optional): Offset to read the value
        """
        self.RootKey = RootKey
        self.OffsetToSubKeyName = OffsetToSubKeyName
        self.OffsetToValueName = OffsetToValueName
        self.OffsetToValue = OffsetToValue
        self.SubKeyName = SubKeyName
        self.ValueName = ValueName
        self.Value = PolicyValue(Value.valueType, Value.value)

    def __eq__(self, other: "Rule") -> bool:
        """Rule table offsets are not considered for equivalency, only the actual key/value."""
        if (
            self.RootKey == other.RootKey
            and self.SubKeyName == other.SubKeyName
            and self.ValueName == other.ValueName
            and self.Value == other.Value
        ):
            return True
        else:
            return False

    @classmethod
    def FromFsAndVtOffset(cls: "Rule", fs: BinaryIO, vtOffset: int) -> "Rule":
        """Construct a Rule initialized from a deserialized stream and vtOffset.

        Args:
            fs (BinaryIO): File stream to read from
            vtOffset (int ): offset in fs to get the ValueTable

        Raises:
            (Exception): Invalid File Stream or value table offset
        """
        if fs is None or vtOffset is None:
            raise Exception("Invalid File stream or Value Table offset")
        (RootKey, OffsetToSubKeyName, OffsetToValueName, OffsetToValue) = struct.unpack(
            cls.StructFormat, fs.read(cls.StructSize)
        )

        orig_offset = fs.tell()
        SubKeyName = PolicyString.FromFileStream(fs=fs, fsOffset=vtOffset + OffsetToSubKeyName)
        ValueName = PolicyString.FromFileStream(fs=fs, fsOffset=vtOffset + OffsetToValueName)
        Value = PolicyValue.FromFileStream(fs=fs, fsOffset=vtOffset + OffsetToValue)
        fs.seek(orig_offset)
        return cls(
            RootKey=RootKey,
            OffsetToSubKeyName=OffsetToSubKeyName,
            OffsetToValueName=OffsetToValueName,
            OffsetToValue=OffsetToValue,
            SubKeyName=SubKeyName,
            ValueName=ValueName,
            Value=Value,
        )

    @classmethod
    def FromFsAndVtBytes(cls: "Rule", fs: BinaryIO, vt: bytes) -> "Rule":
        """Contstruct a Rule initialized from a deserialized stream and value table.

        Args:
            fs (BinaryIO): File stream to read from
            vt (bytes): bytes representing a value table

        Raises:
            (Exception): Invalid File Stream or value table
        """
        if fs is None or vt is None:
            raise Exception("Invalid File stream or Value Table offset")
        (RootKey, OffsetToSubKeyName, OffsetToValueName, OffsetToValue) = struct.unpack(
            cls.StructFormat, fs.read(cls.StructSize)
        )
        SubKeyName = PolicyString.FromBytes(vt, OffsetToSubKeyName)
        ValueName = PolicyString.FromBytes(vt, OffsetToValueName)
        Value = PolicyValue.FromBytes(vt, OffsetToValue)
        return cls(
            RootKey=RootKey,
            OffsetToSubKeyName=OffsetToSubKeyName,
            OffsetToValueName=OffsetToValueName,
            OffsetToValue=OffsetToValue,
            SubKeyName=SubKeyName,
            ValueName=ValueName,
            Value=Value,
        )

    def Print(self, prefix: str = "") -> None:
        """Prints the contents of the Rule.

        Args:
            prefix (:obj:`str`, optional): Prefix to put in front of print statement.
        """
        print("%sRule" % (prefix,))
        print("%s  RootKey:  %x" % (prefix, self.RootKey))
        print("%s  SubKeyNameOffset:  %x" % (prefix, self.OffsetToSubKeyName))
        print("%s  ValueNameOffset:  %x" % (prefix, self.OffsetToValueName))
        print("%s  ValueOffset:  %x" % (prefix, self.OffsetToValue))
        self.SubKeyName.Print(prefix=prefix + "  SubKeyName ")
        self.ValueName.Print(prefix=prefix + "  ValueName ")
        self.Value.Print(prefix=prefix + "  ")

    def Serialize(self, ruleOut: bytearray, valueOut: bytearray, offsetInVT: int) -> None:
        """Serialize Rule.

        Args:
            ruleOut (bytearray): serialized Reserved2 rule
            valueOut (bytearray): not currently used
            offsetInVT (int): not currently used
        """
        self.OffsetToSubKeyName = offsetInVT
        localArray = bytearray()
        self.SubKeyName.Serialize(valueOut=localArray)
        valueOut += localArray

        self.OffsetToValueName = self.OffsetToSubKeyName + len(localArray)
        localArray = bytearray()
        self.ValueName.Serialize(valueOut=localArray)
        valueOut += localArray

        self.OffsetToValue = self.OffsetToValueName + len(localArray)
        localArray = bytearray()
        self.Value.Serialize(valueOut=localArray)
        valueOut += localArray

        ruleOut += struct.pack(
            self.StructFormat, self.RootKey, self.OffsetToSubKeyName, self.OffsetToValueName, self.OffsetToValue
        )


class PolicyValueType:
    """Class for managing PolicyValueTypes.

    Class for storing, serializing, deserializing, & printing PolicyValue Types
    """

    StructFormat = "<H"
    StructSize = struct.calcsize(StructFormat)

    POLICY_VALUE_TYPE_STRING = 0
    POLICY_VALUE_TYPE_BOOL = 1
    POLICY_VALUE_TYPE_DWORD = 2
    POLICY_VALUE_TYPE_DWORD_RANGE = 3
    POLICY_VALUE_TYPE_DWORD_CHOICE = 4
    POLICY_VALUE_TYPE_QWORD = 5
    POLICY_VALUE_TYPE_QWORD_RANGE = 6
    POLICY_VALUE_TYPE_QWORD_CHOICE = 7
    POLICY_VALUE_TYPE_OPTION = 8
    POLICY_VALUE_TYPE_MULTI_STRING = 9
    POLICY_VALUE_TYPE_BINARY = 10

    SupportedValueTypes = {POLICY_VALUE_TYPE_DWORD, POLICY_VALUE_TYPE_QWORD, POLICY_VALUE_TYPE_STRING}

    def __init__(self, Type: int) -> "PolicyValueType":
        """Inits the PolicyValueType.

        Args:
            Type (obj): POLICY_VALUE_TYPE_*

        Raises:
            (Exception): Unsupported ValueType

        """
        if Type not in self.SupportedValueTypes:
            ErrorMessage = "Unsupported ValueType: %x" % Type
            print(ErrorMessage)
            if STRICT:
                raise Exception(ErrorMessage)

        self.vt = Type

    @classmethod
    def FromFileStream(cls: "PolicyValueType", fs: BinaryIO, fsOffset: int = None) -> "PolicyValueType":
        """Load a Policy Value Type from a file stream.

        NOTE: if fsOffset is not specified, stream fs position is at beginning of struct.

        Args:
            fs (BinaryIO): filestream to read
            fsOffset (:obj:`int`, optional): offset to start reading from
        """
        if fsOffset:
            fs.seek(fsOffset)
        valueType = struct.unpack(cls.StructFormat, fs.read(cls.StructSize))[0]
        return cls(valueType)

    @classmethod
    def FromBytes(cls: "PolicyValueType", b: bytes, bOffset: int = 0) -> "PolicyValueType":
        """Inits a PolicyStringValue from a filestream.

        Args:
            b (bytes): bytes to read
            bOffset (:obj:`int`, optional): Offset to start reading from.
        """
        valueType = struct.unpack_from(cls.StructFormat, b, bOffset)[0]
        return cls(valueType)

    def Print(self, prefix: str = "") -> None:
        """Prints the contents of the Policy String Type.

        Args:
            prefix (:obj:`str`, optional): Prefix to put in front of print statement.
        """
        print("%s%s%s" % (prefix, "ValueType: ", self.vt))

    def Serialize(self, valueOut: bytearray) -> None:
        """Serialize the Policy String Type.

        Args:
            valueOut (bytearray): serialized object
        """
        valueOut += struct.pack(self.StructFormat, self.vt)

    def Get(self: "PolicyValueType") -> int:
        """Returns the value type."""
        return self.vt


class PolicyString:
    r"""Class for managing PolicyString structures.

    Class for storing, serializing, deserializing, & printing PolicyString Types
    NOTE: This type is used both in the keys as SubKeyName and ValueName and it
        can be a value.  When used as a value, a NULL may follow the string, but
        the NULL will not be included in the string size

    16-bit, little endian, size of the string in _bytes_
    followed by a UTF-16LE string, no NULL terminator

    Example of "PlatformID"
    \x14\x00P\x00l\x00a\x00t\x00f\x00o\x00r\x00m\x00I\x00D\x00

    The trailing \x00 is not a NULL, it is UTF-16LE encoding of "D"
    """

    StringLengthFormat = "<H"
    StringLengthSize = struct.calcsize(StringLengthFormat)

    def __init__(self, String: str = None) -> "PolicyString":
        """Inits PolicyStructure.

        Args:
            String (:obj:`str`, optional): String to set
        """
        if String:
            self.String = String
        else:
            self.String = ""
        return

    @classmethod
    def FromFileStream(cls: "PolicyString", fs: BinaryIO, fsOffset: int = None) -> "PolicyString":
        """Inits a PolicyString from a file stream.

        Args:
            fs (BinaryIO): File stream to read
            fsOffset (:obj:`int`, optional): Offset to start reading from.

        Raises:
            (Exception): string length mismatch
        """
        if fsOffset:
            fs.seek(fsOffset)
        StringLength = struct.unpack(cls.StringLengthFormat, fs.read(cls.StringLengthSize))[0]
        LocalString = bytes.decode(fs.read(StringLength), encoding="utf_16_le")
        if len(LocalString) != (StringLength / 2):
            raise Exception("String length mismatch")
        return cls(String=LocalString)

    @classmethod
    def FromBytes(cls: "PolicyString", b: bytes, bOffset: int = 0) -> "PolicyString":
        """Inits a PolicyString from a filestream.

        Args:
            b (bytes): bytes to read
            bOffset (:obj:`int`, optional): Offset to start reading from.

        Raises:
            (Exception): string length mismatch
        """
        StringLength = struct.unpack_from(cls.StringLengthFormat, b, bOffset)[0]
        bOffset += struct.calcsize(cls.StringLengthFormat)
        LocalString = bytes.decode(b[bOffset : bOffset + StringLength], encoding="utf_16_le")
        if len(LocalString) != (StringLength / 2):
            raise Exception("String length mismatch")
        return cls(String=LocalString)

    def Print(self, prefix: str = "") -> None:
        """Prints the contents of the Policy String.

        Args:
            prefix (:obj:`str`, optional): Prefix to put in front of print statement.
        """
        print("%s%s%s" % (prefix, "String:  ", self.String))

    def Serialize(self, valueOut: bytearray) -> None:
        """Serialize the Policy String.

        Args:
            valueOut (bytearray): serialized object
        """
        b = str.encode(self.String, encoding="utf_16_le")
        size = struct.pack(self.StringLengthFormat, len(b))
        valueOut += size + b


class PolicyValue:
    """Class for managing PolicyValue structures.

    Class for storing, serializing, deserializing, & printing policy values
    Typically handles primitive types itself, or delegates to other classes
    for non-primitive structures, e.g. PolicyString

    Attributes:
        valueType (PolicyValueType): Policy Value Type
        value (struct): Policy String, dword, qword
    """

    def __init__(self, valueType: PolicyValueType, value: str) -> "PolicyValue":
        """Inits a Policy Value.

        Args:
            valueType (PolicyValueType): Policy Value Type
            value (struct): Policy value string, dword, or qword
        """
        self.valueType = valueType
        self.value = value

    @classmethod
    def FromFileStream(cls: "PolicyValue", fs: BinaryIO, fsOffset: int = None) -> "PolicyValue":
        """Load a Policy Value from a file stream.

        NOTE: if fsOffset is not specified, stream fs position is at beginning of struct.

        Args:
            fs (BinaryIO): filestream to read
            fsOffset (:obj:`int`, optional): offset to start reading from

        Raises:
            (Exception): if STRICT is True, unsupported PolicyValueType
        """
        if fsOffset:
            fs.seek(fsOffset)
        else:
            fsOffset = fs.tell()
        valueType = PolicyValueType.FromFileStream(fs=fs, fsOffset=fsOffset)

        if valueType.Get() is PolicyValueType.POLICY_VALUE_TYPE_STRING:
            value = PolicyString.FromFileStream(fs=fs)
        elif valueType.Get() is PolicyValueType.POLICY_VALUE_TYPE_DWORD:
            value = struct.unpack("<I", fs.read(4))[0]
        elif valueType.Get() is PolicyValueType.POLICY_VALUE_TYPE_QWORD:
            value = struct.unpack("<Q", fs.read(8))[0]
        else:
            value = "Value Type not supported"
            if STRICT:
                raise Exception(value)
        return cls(valueType=valueType, value=value)

    @classmethod
    def FromBytes(cls: "PolicyValue", b: bytes, bOffset: int = 0) -> "PolicyValue":
        """Load a Policy Value from bytes.

        Args:
            b (bytes): bytes to parse
            bOffset (:obj:`int`, optional): offset to start at.
        """
        valueType = PolicyValueType.FromBytes(b, bOffset)
        bOffset += PolicyValueType.StructSize

        if valueType.Get() is PolicyValueType.POLICY_VALUE_TYPE_STRING:
            value = PolicyString.FromBytes(b, bOffset)
        elif valueType.Get() is PolicyValueType.POLICY_VALUE_TYPE_DWORD:
            value = struct.unpack_from("<I", b, bOffset)[0]
        elif valueType.Get() is PolicyValueType.POLICY_VALUE_TYPE_QWORD:
            value = struct.unpack_from("<Q", b, bOffset)[0]
        else:
            value = "Value Type not supported"
            if STRICT:
                raise Exception(value)

        return cls(valueType=valueType, value=value)

    def GetValueType(self) -> str:
        """Returns the value type."""
        return self.valueType

    def Print(self, prefix: str = "") -> None:
        """Prints the contents of the object.

        Args:
            prefix (:obj:`str`, optional): Prefix to put in front of print statement.
        """
        self.valueType.Print(prefix=prefix)
        if isinstance(self.value, int):
            print("%s%s0x%x" % (prefix, "Value:  ", self.value))
        elif self.valueType.vt == PolicyValueType.POLICY_VALUE_TYPE_STRING:
            self.value.Print(prefix + "Value: ")
        else:
            print('%s%s"%s"' % (prefix, "Value:  ", str(self.value)))

    def Serialize(self, valueOut: bytearray) -> None:
        """Serializes the object to valueOut.

        Args:
            valueOut (bytearray): object to write bytes to.

        Raises:
            (Exception): Value Type not supported.
        """
        self.valueType.Serialize(valueOut)

        vt = self.valueType.Get()
        if vt is PolicyValueType.POLICY_VALUE_TYPE_STRING:
            self.value.Serialize(valueOut)
            """ NOTE: add a NULL here for consistency with server-side code """
            valueOut += struct.pack("<H", 0x0000)
        elif vt is PolicyValueType.POLICY_VALUE_TYPE_DWORD:
            valueOut += struct.pack("<I", self.value)
        elif vt is PolicyValueType.POLICY_VALUE_TYPE_QWORD:
            valueOut += struct.pack("<Q", self.value)
        else:
            print("Type not supported")
            if STRICT:
                raise Exception("Value Type not supported")


class Reserved2(object):
    """Class for managing Reserved2 structures.

    For testing non-firmware, legacy policies
    Implementation can do basic parsing of rules but not values
    For test purposes only
    """

    StructFormat = "<III"
    StructSize = struct.calcsize(StructFormat)

    def __init__(self, fs: BinaryIO = None, vtOffset: int = 0) -> "Reserved2":
        """Initializes the Reserved2 structure.

        Args:
            fs (:obj:`BinaryIO`, optional): initialize with a filestream
            vtOffset (:obj:`int`, optional): used if populating from a filestream
        """
        if fs is None:
            self.ObjectType = 0
            self.Element = 0
            self.OffsetToValue = vtOffset
        else:
            self.PopulateFromFileStream(fs, vtOffset)

        errorMessage = "Reserved2 not fully supported"
        if STRICT is True:
            raise Exception(errorMessage)

    def PopulateFromFileStream(self, fs: BinaryIO, vtOffset: int = 0) -> None:
        """Initializes the Reserved2 structure from a filestream.

        Args:
            fs (BinaryIO): file stream to read from
            vtOffset (int): not currently used

        Raises:
            (Exception): Invalid File stream
            (Exception): if STRICT is True, cannot deserialize reserved2 values
        """
        if fs is None:
            raise Exception("Invalid File stream")
        (self.ObjectType, self.Element, self.OffsetToValue) = struct.unpack(self.StructFormat, fs.read(self.StructSize))

        errorMessage = "Reserved2 PopulateFromFileStream does not deserialize Reserved2 values"
        print(errorMessage)
        if STRICT is True:
            raise Exception(errorMessage)

    def Print(self, prefix: str = "") -> None:
        """Prints the Reserved2 structure.

        Args:
            prefix (str): prefix to append when printing to terminal
        """
        print("%sReserved2" % prefix)
        print("%s  ObjectType:  %x" % (prefix, self.ObjectType))
        print("%s  Element:  %x" % (prefix, self.Element))
        print("%s  ValueOffset:  %x" % (prefix, self.OffsetToValue))

    def Serialize(self, ruleOut: bytearray, valueOut: bytearray = None, offsetInVT: int = 0) -> None:
        """Serialize Reserved2 rule.

        Args:
            ruleOut (bytearray): serialized Reserved2 rule
            valueOut (bytearray): not currently used
            offsetInVT (int): not currently used

        Raises:
            (Exception): if STRICT is True, value serialization not supported
        """
        ruleOut += struct.pack(self.StructFormat, self.ObjectType, self.Element, self.OffsetToValue)
        errorMessage = "Reserved2 value serialization not supported"
        print(errorMessage)
        if STRICT is True:
            raise Exception(errorMessage)


class FirmwarePolicy(object):
    """Class for managing Firmware Policy structures.

    Manages storing, serializing, deserializing, & printing Firmware Policy structures

    Typically handles primitive types itself, or delegates to other classes
    for non-primitive structures, e.g. Rule, PolicyValue, PolicyString
    """

    FixedStructFormat = "<HI16sHIHH"  # omits completely unsupported Reserved1 array
    FixedStructSize = struct.calcsize(FixedStructFormat)

    POLICY_BLOB_MIN_SIZE = 32  # bytes
    POLICY_FORMAT_VERSION = 2
    POLICY_VERSION = 1
    POLICY_PUBLISHER = uuid.UUID("5AE6F808-8384-4EB9-A23A-0CCC1093E3DD")  # Do NOT change
    FW_POLICY_ROOT_KEY = 0xEF100000
    FW_POLICY_SUB_KEY_NAME_TARGET = "Target"
    FW_POLICY_SUB_KEY_NAME = "UEFI"
    FW_POLICY_VALUE_NAME = "Policy"
    FW_POLICY_TYPE = PolicyValueType.POLICY_VALUE_TYPE_QWORD
    FW_POLICY_VALUE_DEFINED_MASK = 0x00000000FFFFFFFF
    FW_POLICY_VALUE_OEM_MASK = 0xFFFFFFFF00000000
    FW_POLICY_VALUE_ACTIONS_MASK = 0x0000FFFF0000FFFF
    FW_POLICY_VALUE_STATES_MASK = 0xFFFF0000FFFF0000

    """Defined Policy Actions"""
    FW_POLICY_VALUE_ACTION_SECUREBOOT_CLEAR = 0x0000000000000001
    FW_POLICY_VALUE_ACTION_TPM_CLEAR = 0x0000000000000002

    FW_POLICY_VALUE_ACTION_STRINGS = {
        FW_POLICY_VALUE_ACTION_SECUREBOOT_CLEAR: "Clear UEFI Secure Boot Keys",
        FW_POLICY_VALUE_ACTION_TPM_CLEAR: "Clear the Trusted Platform Module (TPM)",
    }

    """Defined Policy States"""
    FW_POLICY_VALUE_STATE_TBD = 0x0000000000010000

    FW_POLICY_VALUE_STATE_STRINGS = {FW_POLICY_VALUE_STATE_TBD: "To Be Defined Placeholder"}

    def __init__(self, fs: BinaryIO = None) -> "FirmwarePolicy":
        """Initializes a Firmware Policy object.

        Args:
            fs (:obj:`BinaryIO`, optional): Initializes from a Filestream.
        """
        if fs is None:
            self.FormatVersion = self.POLICY_FORMAT_VERSION
            self.PolicyVersion = self.POLICY_VERSION
            self.PolicyPublisher = self.POLICY_PUBLISHER
            self.Reserved1Count = 0
            self.Reserved1 = []
            self.OptionFlags = 0
            self.Reserved2Count = 0
            self.RulesCount = 0
            self.Reserved2 = []
            self.Rules = []
            self.ValueTableSize = 0
            self.ValueTableOffset = 0
            self.ValueTable = []
            self.ValueTableFromFile = None
            self.parseValueTableViaBytes = True
        else:
            self.FromFileStream(fs)

    def AddRule(self, regRule: Rule) -> bool:
        """Adds a rule to the Firmware Policy object representation.

        WARNING: Does not update the valuetable, use Serialize to do that after all rules are added.

        Args:
            regRule (Rule): rule to add

        Returns:
            (bool): True if added, False if it already existed.
        """
        for rule in self.Rules:
            if rule == regRule:
                return False
        self.Rules.append(regRule)
        self.RulesCount += 1
        return True

    def SetDevicePolicy(self, policy: int) -> None:
        """Adds a Rule for the 64-bit policy value bitfield.

        NOTE: The "key" is a well-known constant assigned in the body of this method

        Args:
            policy (int): the 64-bit bitfield value for associated key.
        """
        policyVT = PolicyValueType(Type=PolicyValueType.POLICY_VALUE_TYPE_QWORD)
        SubKeyName = PolicyString(String=self.FW_POLICY_SUB_KEY_NAME)
        ValueName = PolicyString(String=self.FW_POLICY_VALUE_NAME)
        Value = PolicyValue(valueType=policyVT, value=policy)
        rule = Rule(RootKey=self.FW_POLICY_ROOT_KEY, SubKeyName=SubKeyName, ValueName=ValueName, Value=Value)
        self.AddRule(rule)

    def SetDeviceTarget(self, target: dict) -> None:
        """Sets the device target dictionary.

        Args:
            target (dict): ValueName/Value pairs
        """
        for k, v in target.items():
            ValueName = PolicyString(String=k)
            if k == "Nonce":
                policyVT = PolicyValueType(Type=PolicyValueType.POLICY_VALUE_TYPE_QWORD)
                Value = PolicyValue(valueType=policyVT, value=v)
            else:
                policyVT = PolicyValueType(Type=PolicyValueType.POLICY_VALUE_TYPE_STRING)
                Value = PolicyValue(valueType=policyVT, value=PolicyString(String=v))
            rule = Rule(
                RootKey=self.FW_POLICY_ROOT_KEY,
                SubKeyName=PolicyString(String=self.FW_POLICY_SUB_KEY_NAME_TARGET),
                ValueName=ValueName,
                Value=Value,
            )
            self.AddRule(rule)

    def SerializeToStream(self, stream: BinaryIO) -> None:
        """Serializes the Firmware Policy to a stream.

        Args:
            stream (BinaryIO): stream to write to
        """
        ba = bytearray()
        self.Serialize(output=ba)
        stream.write(ba)

    def Serialize(self, output: bytearray) -> None:
        """Serializes the Firmware Policy to a bytearray.

        Args:
            output (bytearray): bytearray to write to.
        """
        if self.Reserved1Count > 0:
            ErrorMessage = "Reserved1 not supported"
            if STRICT is True:
                raise Exception(ErrorMessage)
            print(ErrorMessage)

        fixedSizeHeader = struct.pack(
            self.FixedStructFormat,
            self.FormatVersion,
            self.PolicyVersion,
            self.PolicyPublisher.bytes_le,
            self.Reserved1Count,
            self.OptionFlags,
            self.Reserved2Count,
            self.RulesCount,
        )

        Reserved2Offset = len(fixedSizeHeader)
        Reserved2Size = self.Reserved2Count * Reserved2.StructSize
        RulesOffset = Reserved2Offset + Reserved2Size
        RulesSize = self.RulesCount * Rule.StructSize
        self.ValueTableOffset = RulesOffset + RulesSize

        offsetInVT = 0
        ruleArray = bytearray()
        valueArray = bytearray()

        for i in range(self.Reserved2Count):
            rule = bytearray()
            value = bytearray()
            self.Reserved2[i].Serialize(ruleOut=rule, valueOut=value, offsetInVT=offsetInVT)
            ruleArray += rule
            valueArray += value
            offsetInVT += len(value)

        for i in range(self.RulesCount):
            rule = bytearray()
            value = bytearray()
            self.Rules[i].Serialize(ruleOut=rule, valueOut=value, offsetInVT=offsetInVT)
            ruleArray += rule
            valueArray += value
            offsetInVT += len(value)

        serial = bytearray(fixedSizeHeader)
        serial += ruleArray

        self.ValueTableOffset = len(serial)
        self.ValueTableSize = len(valueArray)
        self.ValueTable = valueArray

        serial += valueArray
        output += serial
        self.ValueTableFromFile = False

    def FromFileStream(self, fs: BinaryIO, parseByBytes: bool = True) -> None:
        """Initializes the Firmware Policy from a FileStream.

        Args:
            fs (BinaryIO): Filestream to read from
            parseByBytes (:obj:`bool`, optional): whether to parse the value table by bytes or offset

        Raises:
            (Exception): an invalid file stream was provided.
        """
        if fs is None:
            raise Exception("Invalid File stream")

        self.parseValueTableViaBytes = parseByBytes

        begin = fs.tell()
        fs.seek(0, io.SEEK_END)
        end = fs.tell()  # end is offset after last byte
        fs.seek(begin)
        size = end - begin

        if size < self.POLICY_BLOB_MIN_SIZE:
            raise Exception("Policy is too small")

        self.FormatVersion = struct.unpack("<H", fs.read(2))[0]
        if self.FormatVersion > self.POLICY_FORMAT_VERSION:
            print("Policy Format Version %x is not supported" % self.FormatVersion)
            raise Exception("Policy Format Version is newer than supported")

        self.PolicyVersion = struct.unpack("<I", fs.read(4))[0]
        PolicyPublisher = struct.unpack("<16s", fs.read(struct.calcsize("<16s")))[0]
        self.PolicyPublisher = uuid.UUID(bytes_le=PolicyPublisher)

        self.Reserved1Count = struct.unpack("<H", fs.read(2))[0]
        if STRICT and (self.Reserved1Count > 0):
            raise Exception("Reserved1 not supported")
        self.Reserved1 = []
        for i in range(self.Reserved1Count):
            Reserved1 = struct.unpack("<16s", fs.read(struct.calcsize("<16s")))[0]
            self.Reserved1.append(uuid.UUID(bytes_le=Reserved1))

        self.OptionFlags = struct.unpack("<I", fs.read(4))[0]

        self.Reserved2Count = struct.unpack("<H", fs.read(2))[0]
        self.RulesCount = struct.unpack("<H", fs.read(2))[0]

        # now we pause our parsing to bounds check the variable size structures
        Reserved2Offset = fs.tell()
        Reserved2Size = self.Reserved2Count * Reserved2.StructSize
        if (Reserved2Offset + Reserved2Size) > end:
            raise Exception("Reserved2 larger than buffer")

        RulesOffset = Reserved2Offset + Reserved2Size
        RulesSize = self.RulesCount * Rule.StructSize
        if (RulesOffset + RulesSize) > end:
            raise Exception("Rules larger than buffer")

        self.ValueTableOffset = RulesOffset + RulesSize
        self.ValueTableSize = end - self.ValueTableOffset
        saved_fs = fs.tell()
        fs.seek(self.ValueTableOffset)
        self.ValueTable = bytes(fs.read(self.ValueTableSize))
        self.ValueTableFromFile = True
        fs.seek(saved_fs)

        # resume parsing the variable length structures using table offset
        self.Reserved2 = []
        for i in range(self.Reserved2Count):
            self.Reserved2.append(Reserved2(fs=fs, vtOffset=self.ValueTableOffset))

        self.Rules = []
        for i in range(self.RulesCount):
            if self.parseValueTableViaBytes is True:
                RegRule = Rule.FromFsAndVtBytes(fs=fs, vt=self.ValueTable)
            else:
                RegRule = Rule.FromFsAndVtOffset(fs=fs, vtOffset=self.ValueTableOffset)
            self.Rules.append(RegRule)

    def PrintDevicePolicy(self, devicePolicy: int, prefix: str = "") -> None:
        """Prints the device policy."""
        prefix = prefix + "    "
        for bit in self.FW_POLICY_VALUE_ACTION_STRINGS.keys():
            if (devicePolicy & bit) != 0:
                print(prefix + self.FW_POLICY_VALUE_ACTION_STRINGS[bit])

    def Print(self) -> None:
        """Prints the firmware policy structure."""
        prefix = "  "
        print("Firmware Policy")
        print("  FormatVersion:    %x" % self.FormatVersion)
        print("  PolicyVersion:    %x" % self.PolicyVersion)
        print("  PolicyPublisher:  %s" % self.PolicyPublisher)
        print("  Reserved1Count:   %x" % self.Reserved1Count)
        for item in self.Reserved1:
            print("   Reserved1:        %s" % item)
        print("  OptionFlags:  %x" % self.OptionFlags)
        print("  Reserved2Count:  %x" % self.Reserved2Count)
        print("  RulesCount:  %x" % self.RulesCount)
        print("  ValueTableSize:  %x" % self.ValueTableSize)
        print("  ValueTableOffset:  %x" % self.ValueTableOffset)
        for rule in self.Reserved2:
            rule.Print(prefix=prefix)
        for rule in self.Rules:
            rule.Print(prefix=prefix)
            if (
                rule.RootKey == self.FW_POLICY_ROOT_KEY
                and rule.SubKeyName.String == self.FW_POLICY_SUB_KEY_NAME
                and rule.ValueName.String == self.FW_POLICY_VALUE_NAME
            ):
                print(prefix + "  Device Policy:")
                self.PrintDevicePolicy(devicePolicy=rule.Value.value, prefix=prefix)
        print("  Valuetable")
        print(self.ValueTable)
        return



# @file
# Code to help convert an Int to StatusCode string
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module for converting an Int to StatusCode string."""


class UefiStatusCode(object):
    """Object representing a UEFI Status Code from Appendix D of the UEFI spec."""

    # high bit set
    ErrorCodeStrings = [
        "NOT VALID",
        "Load Error",
        "Invalid Parameter",
        "Unsupported",
        "Bad BufferSize",
        "Buffer Too Small",
        "Not Ready",
        "Device Error",
        "Write Protected",
        "Out of Resources",
        "Volume Corrupt",
        "Volume Full",
        "No Media",
        "Media Changed",
        "Not Found",
        "Access Denied",
        "No Response",
        "No Mapping",
        "Time Out",
        "Not Started",
        "Already Started",
        "Aborted",
        "ICMP Error",
        "TFTP Error",
        "Protocol Error",
        "Incompatible Error",
        "Security Violation",
        "CRC Error",
        "End of Media",
        "Reserved(29)",
        "Reserved(30)",
        "End of File",
        "Invalid Language",
        "Compromised Data",
        "IP Address Conflict",
        "HTTP Error",
    ]

    NonErrorCodeStrings = [
        "Success",
        "Unknown Glyph",
        "Delete Failure",
        "Write Failure",
        "Buffer Too Small",
        "Stale Data",
        "File System",
        "Reset Required",
    ]

    def Convert32BitToString(self, value: int) -> str:
        """Convert 32 bit int to a friendly UEFI status code string value."""
        StatusStrings = UefiStatusCode.NonErrorCodeStrings

        if (value >> 31) & 1 == 1:
            # error
            StatusStrings = UefiStatusCode.ErrorCodeStrings
            value = value & 0x7FFFFFFF  # mask off upper bit

        if value >= len(StatusStrings):
            return "Undefined StatusCode"

        return StatusStrings[value]

    def Convert64BitToString(self, value: int) -> str:
        """Convert 64 bit int to a friendly UEFI status code string value."""
        StatusStrings = UefiStatusCode.NonErrorCodeStrings

        if (value >> 63) & 1 == 1:
            # error
            StatusStrings = UefiStatusCode.ErrorCodeStrings
            value = value & 0x7FFFFFFFFFFFFFFF  # mask off upper bit

        if value >= len(StatusStrings):
            return "Undefined StatusCode"

        return StatusStrings[value]

    def ConvertHexString64ToString(self, hexstring: str) -> str:
        """Convert 64 bit hexstring in 0x format to a UEFI status code."""
        value = int(hexstring, 16)
        return self.Convert64BitToString(value)

    def ConvertHexString32ToString(self, hexstring: str) -> str:
        """Convert 32 bit hexstring in 0x format to a UEFI status code."""
        value = int(hexstring, 16)
        return self.Convert32BitToString(value)



##
# File to mark this a python package
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""This package contains different EDK2 file generators for capsules."""



# @file
# Module contains helper classes and functions to work with UEFI FFs.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module containing helper classes and functions for working with UEFI FFs."""

import struct
import sys
import uuid
from typing import IO


class EfiFirmwareFileSystemHeader(object):
    """An object representing an EFI_FFS_FILE_HEADER.

    Can parse or produce an EFI_FFS_FILE_HEADER structure/byte buffer.

    ```
    typedef struct {
        EFI_GUID                Name;
        EFI_FFS_INTEGRITY_CHECK IntegrityCheck;
        EFI_FV_FILETYPE         Type;
        EFI_FFS_FILE_ATTRIBUTES Attributes;
        UINT8                   Size[3];
        EFI_FFS_FILE_STATE      State;
    } EFI_FFS_FILE_HEADER;
    ```
    """

    def __init__(self) -> "EfiFirmwareFileSystemHeader":
        """Inits an empty object."""
        self.StructString = "=16sHBBBBBB"  # spell-checker: disable-line
        self.FileSystemGuid = None
        self.Size0 = None
        self.Size1 = None
        self.Size2 = None
        self.Attributes = None
        self.Type = None
        self.State = None

    def get_size(self) -> int:
        """Returns the size of the header."""
        return self.Size0 + (self.Size1 << 8) + (self.Size2 << 16)

    def load_from_file(self, file: IO) -> "EfiFirmwareFileSystemHeader":
        """Loads data into the object from a filestream.

        Args:
            file (obj): An open file that has been seeked to the correct location.

        Returns:
            (EfiFirmwareFileSystemHeader): self
        """
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (
            self.FileSystemGuid,
            self.Checksum,
            self.Type,
            self.Attributes,
            self.Size0,
            self.Size1,
            self.Size2,
            self.State,
        ) = struct.unpack(self.StructString, struct_bytes)

        # Update the GUID to be a UUID object.
        if sys.byteorder == "big":
            self.FileSystemGuid = uuid.UUID(bytes=self.FileSystemGuid)
        else:
            self.FileSystemGuid = uuid.UUID(bytes_le=self.FileSystemGuid)

        return self

    def serialize(self) -> bytes:
        r"""Serializes the object.

        Returns:
            (bytes): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        file_system_guid_bin = self.FileSystemGuid.bytes if sys.byteorder == "big" else self.FileSystemGuid.bytes_le
        return struct.pack(
            self.StructString,
            file_system_guid_bin,
            self.Checksum,
            self.Type,
            self.Attributes,
            self.Size0,
            self.Size1,
            self.Size2,
            self.State,
        )
